import logging
import argparse
import jax
import jax.numpy as jnp

import tinygp
import smolgp
from benchmark import run_benchmark, run_pred_benchmark
from benchmark import save_benchmark_data

from funcs import ss_llh, qs_llh, gp_llh, pss_llh
from funcs import ss_cond, qs_cond, gp_cond, pss_cond
from funcs import ss_pred, qs_pred, gp_pred
from funcs import iss_llh, igp_llh, ipss_llh
from funcs import iss_cond, igp_cond, ipss_cond
from funcs import iss_pred, igp_pred

key = jax.random.PRNGKey(0)
jax.config.update("jax_enable_x64", True)

# Suppress only JAX XLA bridge warnings
logging.getLogger("jax._src.xla_bridge").setLevel(logging.ERROR)

######################################## MAIN ########################################
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Benchmark smolgp/tinygp")
    parser.add_argument(
        "func", type=str, help="Function to benchmark: 'llh' or 'cond'."
    )
    parser.add_argument("--gpu", action="store_true", help="Run on GPU (default: CPU).")
    parser.add_argument(
        "--int",
        action="store_true",
        help="Run with integrated data (default: instantaneous data).",
    )
    args = parser.parse_args()

    # # Set device
    if args.gpu:
        # jax.config.update("jax_platform_name", "gpu")
        print("Running benchmark on GPU")
        machine = "gpu"
        if args.int:
            cutoffs = {"GP": 0, "SSM": 0, "QSM": 0, "pSSM": 1e6}
        else:
            cutoffs = {"GP": 0, "SSM": 0, "QSM": 0, "pSSM": 1e7}
    else:
        # jax.config.update("jax_platform_name", "cpu")
        print("Running benchmark on CPU")
        machine = "cpu"
        cutoffs = {"GP": 6e4, "SSM": 1e7, "QSM": 1e7, "pSSM": 0}

    ## Setup function dictionaries
    llh_funcs = [
        {"SSM": ss_llh, "QSM": qs_llh, "GP": gp_llh, "pSSM": pss_llh},
        {"SSM": iss_llh, "GP": igp_llh, "pSSM": ipss_llh},
    ]
    cond_funcs = [
        {"SSM": ss_cond, "QSM": qs_cond, "GP": gp_cond, "pSSM": pss_cond},
        {"SSM": iss_cond, "GP": igp_cond, "pSSM": ipss_cond},
    ]
    pred_funcs = [
        {"SSM": ss_pred, "QSM": qs_pred, "GP": gp_pred},
        {"SSM": iss_pred, "GP": igp_pred},
    ]
    ################### True GP parameters ######################
    S = 2.36
    w = 0.0195
    Q = 7.63
    sigma = jnp.sqrt(S * w * Q)
    true_kernel = tinygp.kernels.quasisep.SHO(omega=w, quality=Q, sigma=sigma)
    ################# Which kernels to benchmark ##################
    if args.int:
        ssm_kernel = smolgp.kernels.integrated.IntegratedSHO(
            omega=w, quality=Q, sigma=sigma, num_inst=1
        )
        gp_kernel = smolgp.kernels.dense.IntegratedSHOKernel(w=w, Q=Q, S=S)
        kernels = {
            "SSM": ssm_kernel,
            "GP": gp_kernel,
            "pSSM": ssm_kernel,
        }
    else:
        qsm_kernel = tinygp.kernels.quasisep.SHO(omega=w, quality=Q, sigma=sigma)
        ssm_kernel = smolgp.kernels.SHO(omega=w, quality=Q, sigma=sigma)
        gp_kernel = smolgp.kernels.dense.SHOKernel(w=w, Q=Q, S=S)
        kernels = {
            "SSM": ssm_kernel,
            "QSM": qsm_kernel,
            "GP": gp_kernel,
            "pSSM": ssm_kernel,
        }
    ################ Data properties ####################
    yerr = 0.3
    texp = 140.0 if args.int else 0.0
    readout = 40.0 if args.int else 0.0
    if args.int:
        print("Using integrated data with texp =", texp, "and readout =", readout)
    ############################################################
    if args.func in ["llh", "cond"]:
        if args.func == "llh":
            print("Benchmarking likelihood...")
            funcs = llh_funcs[int(args.int)]
            n_repeat = 7
            N_N = 17
            logN_min = 1
            logN_max = 7
        elif args.func == "cond":
            print("Benchmarking condition...")
            funcs = cond_funcs[int(args.int)]
            n_repeat = 7
            N_N = 17
            logN_min = 1
            logN_max = 7

        Ns, runtime, memory, outputs = run_benchmark(
            true_kernel,
            funcs,
            kernels,
            yerr=yerr,
            n_repeat=n_repeat,
            N_N=N_N,
            logN_min=logN_min,
            logN_max=logN_max,
            cutoffs=cutoffs,
            drop_outliers=True,
            use_gpu_profiler=args.gpu,
            exposure_quantities=(texp, readout) if args.int else None,
        )
    elif args.func == "pred":
        print("Benchmarking prediction...")
        funcs = pred_funcs[int(args.int)]
        if args.gpu:
            # cutoffs={"GP": 3e5, "SSM": 3e5, "QSM": 3e5, "pSSM": 3e5} # these are now cutoffs in M
            cutoffs = {"GP": 0, "SSM": 1e6, "QSM": 0}  # these are now cutoffs in M
        else:
            cutoffs = {"GP": 1e6, "SSM": 1e6, "QSM": 1e6}  # these are now cutoffs in M

        # M is set to be 100x N inside run_pred_benchmark
        Ns, runtime, memory, outputs = run_pred_benchmark(
            true_kernel,
            funcs,
            kernels,
            yerr=yerr,
            n_repeat=7,
            N_N=17,
            logN_min=1,
            logN_max=7,
            maxN=1e5,  # in N
            cutoffs=cutoffs,  # in M
            use_gpu_profiler=args.gpu,
            exposure_quantities=(texp, readout) if args.int else None,
        )
    else:
        raise ValueError("Argument must be one of 'llh', 'cond', or 'pred'.")

    isinst = "_int" if args.int else ""
    out_filename = f"results/{machine}_{args.func}{isinst}_benchmark.pkl"
    print("Wrote results to", out_filename)
    save_benchmark_data(out_filename, Ns, runtime, memory, outputs)



import jax
import jax.numpy as jnp
import tinygp
import smolgp

from tests.utils import generate_integrated_data
from tests.test_kernels import (
    kernel_function,
    likelihood,
    condition,
    predict,
)

key = jax.random.PRNGKey(0)
jax.config.update("jax_enable_x64", True)


def test_integrated():
    ## SHO Kernel
    S = 2.36
    w = 0.0195
    Q = 7.63
    sigma = jnp.sqrt(S * w * Q)

    ## Instantaneous kernel for generating data
    true_kernel = tinygp.kernels.quasisep.SHO(omega=w, quality=Q, sigma=sigma)

    ## Generate mock data
    Ninst = 2  # 3
    Ns = [30, 50, 80]
    yerr = [0.3, 0.6, 0.24]
    texps = [140.0, 55, 12]
    readouts = [40.0, 28.0, 40.0]
    t_train = []
    y_train = []
    texp_train = []
    yerr_train = []
    instid = []
    for n in range(Ninst):
        t, y = generate_integrated_data(
            Ns[n], true_kernel, texp=texps[n], readout=readouts[n], yerr=yerr[n]
        )
        t_train.append(t)
        y_train.append(y)
        texp_train.append(jnp.full_like(t, texps[n]))
        yerr_train.append(jnp.full_like(t, yerr[n]))
        instid.append(jnp.full_like(t, n).astype(int))  # has to be integer
    t_train = jnp.concatenate(t_train)
    y_train = jnp.concatenate(y_train)
    texp_train = jnp.concatenate(texp_train)
    yerr_train = jnp.concatenate(yerr_train)
    instid = jnp.concatenate(instid)
    X_train = (t_train, texp_train, instid)

    # Integrated kernels
    kernel_smol = smolgp.kernels.integrated.IntegratedSHO(
        omega=w, quality=Q, sigma=sigma, num_insts=Ninst
    )
    kernel_tiny = smolgp.kernels.dense.IntegratedSHOKernel(S=S, w=w, Q=Q)

    print("Testing IntegratedSHO kernel...")

    # Test k(Delta) agrees
    kernel_function(kernel_smol, kernel_tiny, tol=1e-9, atol=1e-12)

    # Build GP objects
    gp_smol = smolgp.GaussianProcess(kernel=kernel_smol, X=X_train, diag=yerr_train**2)
    gp_tiny = tinygp.GaussianProcess(kernel=kernel_tiny, X=X_train, diag=yerr_train**2)

    # Check likelihood
    likelihood(gp_smol, gp_tiny, y_train, tol=1e-10, atol=1e-13)

    # Check conditioning
    condition(gp_smol, gp_tiny, y_train, tol=1e-9, atol=1e-12)

    # Check predictions
    predict(gp_smol, gp_tiny, y_train, tol=1e-9, atol=1e-12)

    # TODO: test predict with exposure times


if __name__ == "__main__":
    test_integrated()
    print("All integrated kernel tests passed.")



from __future__ import annotations

import jax
import jax.numpy as jnp


def ParallelIntegratedRTSSmoother(
    kernel,
    t_states,
    stateid,
    instid,
    kalman_results,
):
    """
    Wrapper for Parallel RTS smoother

    Parameters:
        kernel: StateSpaceModel kernel
        t_states: time coordinates of the states
        stateid: exposure start/end indicators
        instid: instrument IDs
        kalman_results: output from Kalman filter:
            m_pred, P_pred, m_filter, P_filter

    Returns:
        E:
        g: smoothed means
        L: smoothed covariances
    """
    m_pred, P_pred, m_filter, P_filter = kalman_results
    Phi_aug = kernel.transition_matrix
    Q_aug = kernel.process_noise
    RESET = kernel.reset_matrix

    asso_params = make_associative_params(
        Phi_aug,
        Q_aug,
        RESET,
        t_states,
        stateid,
        instid,
        m_pred,
        P_pred,
        m_filter,
        P_filter,
    )
    E, g, L = parallel_integrated_rts_smoother(asso_params)
    m_smoothed = g
    P_smoothed = L
    return m_smoothed, P_smoothed


@jax.jit
def make_associative_params(
    Phi_aug,
    Q_aug,
    RESET,
    t_states,
    stateid,
    instid,
    m_pred,
    P_pred,
    m_filter,
    P_filter,
):
    """Generate the associative parameters needed for parallel RTS

    See eqns in Section 4B of Sarkka & Garcia-Fernandez (2020)
    """

    def make_last_params(mf_last, Pf_last):
        return (jnp.zeros_like(Pf_last), mf_last, Pf_last)

    def make_generic_params(
        Phi_dt,
        Q_dt,
        Reset,
        mp,
        Pp,
        mf,
        Pf,
        sid_current,
    ):
        def end_state():
            Phik = Phi_dt
            Qk = Q_dt
            mk = mf
            Pk = Pf
            return Phik, Qk, mk, Pk

        def start_state():
            Phik = Phi_dt @ Reset
            Qk = Q_dt
            mk = mp
            Pk = Pp
            return Phik, Qk, mk, Pk

        Phik, Qk, mk, Pk = jax.lax.cond(
            sid_current == 0,
            start_state,
            end_state,
        )

        A = Phik @ Pk @ Phik.T + Qk
        b = Pk @ Phik.T

        E = jax.scipy.linalg.solve(A.T, b.T, assume_a="pos").T
        g = mk - E @ (Phik @ mk)
        L = Pk - E @ Phik @ Pk

        return (E, g, L)

    t_delta = jnp.diff(t_states)
    Phis = jax.vmap(Phi_aug, in_axes=(None, 0))(0, t_delta)
    Qs = jax.vmap(Q_aug, in_axes=(None, 0))(0, t_delta)
    Resets = jax.vmap(RESET)(instid[:-1])

    E, g, L = jax.vmap(
        make_generic_params,
    )(
        Phis,
        Qs,
        Resets,
        m_pred[:-1],
        P_pred[:-1],
        m_filter[:-1],
        P_filter[:-1],
        stateid[:-1],
    )

    EN, gN, LN = make_last_params(m_filter[-1], P_filter[-1])

    E_all = jnp.concatenate([E, EN[jnp.newaxis, ...]], axis=0)
    g_all = jnp.concatenate([g, gN[jnp.newaxis, ...]], axis=0)
    L_all = jnp.concatenate([L, LN[jnp.newaxis, ...]], axis=0)

    return (E_all, g_all, L_all)


def _combine_per_pair(left, right):
    Ei, gi, Li = left
    Ej, gj, Lj = right

    # The indices need to be swapped for some reason...
    Eij = Ej @ Ei
    gij = Ej @ gi + gj
    Lij = Ej @ Li @ Ej.T + Lj

    return (Eij, gij, Lij)


@jax.jit
def parallel_integrated_rts_smoother(asso_params):
    """
    Jax implementation of the parallel RTS smoother algorithm
    for integrated measurements.

    See Section 4B of Sarkka & Garcia-Fernandez (2020) for
    a detailed description of the algorithm and notation,
    and Section 3.2.4 of Rubenzahl & Hattori et al. (2025)
    for the integrated case.

    Total runtime (span) complexity is O(N/T + logT) where N is the
    number of time steps and T is the number of parallel threads.
    """
    return jax.lax.associative_scan(
        jax.vmap(_combine_per_pair),
        asso_params,
        reverse=True,
    )



from __future__ import annotations

import jax
import jax.numpy as jnp


def IntegratedRTSSmoother(kernel, t_states, obsid, instid, stateid, kalman_results):
    """
    Wrapper for jitted integrated_rts_smoother function

    Parameters:
        kernel  : IntegratedStateSpaceModel kernel
        t_states: Array of size K, sorted time coordinate of all states (exposure starts and ends)
        obsid   : Array of size N, which observation (0,...,N-1) is being made at each state k
        instid  : Array of size N, which instrument (0,...,Ninst-1) recorded observation n
        stateid : Array of size K, 0 for exposure-start, 1 for exposure-end
        kalman_results: output from Kalman filter (m_filtered, P_filtered, m_predicted, P_predicted)

    Returns:
        m_filtered: filtered means
        P_filtered: filtered covariances
        m_predicted: predicted means
        P_predicted: predicted covariances
    """

    # Model components
    A_aug = kernel.transition_matrix
    RESET = kernel.reset_matrix

    return integrated_rts_smoother(A_aug, RESET, t_states, obsid, instid, stateid, *kalman_results)


@jax.jit
def integrated_rts_smoother(
    A_aug,
    RESET,
    t_states,
    obsid,
    instid,
    stateid,
    m_filtered,
    P_filtered,
    m_predicted,
    P_predicted,
):
    """
    Jax implementation of the integrated RTS smoothing algorithm

    See Section 3.2.2 in Rubenzahl & Hattori et al. (in prep)
    for detailed description of the algorithm and notation.
    """

    def step(carry, k):
        # Outputs from Kalman filter, unpacked for notational consistency
        m_k = m_filtered[k]
        P_k = P_filtered[k]
        m_pred_next = m_predicted[k + 1]  # has superscript minus
        P_pred_next = P_predicted[k + 1]  # has superscript minus

        # Unpack state and covariance from last iteration
        m_hat_next, P_hat_next = carry

        # Compute smoothing gain
        Delta = t_states[k + 1] - t_states[k]
        A_k = A_aug(0, Delta)

        def smooth_start():
            """RTS smooth an exposure-start state"""

            m_k_pre = m_predicted[k]  # pre-reset start state
            P_k_pre = P_predicted[k]  # pre-reset start covariance

            Reset = RESET(instid[obsid[k]])
            AR = A_k @ Reset
            G_k = jnp.linalg.solve(P_pred_next.T, (P_k_pre @ AR.T).T).T
            m_hat_k = m_k_pre + G_k @ (m_hat_next - m_pred_next)
            P_hat_k = P_k_pre + G_k @ (P_hat_next - P_pred_next) @ G_k.T
            return m_hat_k, P_hat_k

        def smooth_end():
            """RTS smooth an exposure-end state"""
            G_k = jnp.linalg.solve(P_pred_next.T, (P_k @ A_k.T).T).T
            m_hat_k = m_k + G_k @ (m_hat_next - m_pred_next)
            P_hat_k = P_k + G_k @ (P_hat_next - P_pred_next) @ G_k.T
            return m_hat_k, P_hat_k

        m_hat_k, P_hat_k = jax.lax.cond(
            stateid[k] == 0,
            lambda _: smooth_start(),
            lambda _: smooth_end(),
            operand=None,
        )

        return (m_hat_k, P_hat_k), (m_hat_k, P_hat_k)

    # Start smoothing from final filtered state
    init_carry = (m_filtered[-1], P_filtered[-1])

    # Run backward from N-2 down to 0
    K = len(t_states)  # number of iterations
    _, outputs = jax.lax.scan(step, init_carry, jnp.arange(K - 2, -1, -1))
    m_smooth_reversed, P_smooth_reversed = outputs

    # Reverse outputs (with final filtered=smoothed state) to match time order
    m_smooth = jnp.vstack([m_smooth_reversed[::-1], m_filtered[-1][None, :]])
    P_smooth = jnp.vstack([P_smooth_reversed[::-1], P_filtered[-1][None, :, :]])
    return m_smooth, P_smooth



r"""
This subpackage implements the Kalman filter and RTS smoother for an augmented
state space model which includes an integral state as a
:class:`IntegratedStateSpaceSolver`. This is intended to be used with
:class:`smolgp.kernels.integrated` state space models to properly account
for integrated (e.g. exposure-averaged) measurements.

TODO: add documentation from the paper as a refresher/explainer of the augmented approach
        see this file in tinygp for example doc style

"""

from smolgp.solvers.integrated.solver import (
    IntegratedStateSpaceSolver as IntegratedStateSpaceSolver,
)
from smolgp.solvers.integrated.parallel.solver import (
    ParallelIntegratedStateSpaceSolver as ParallelIntegratedStateSpaceSolver,
)



"""
This module uses tinygp to construct dense representations of
kernels for testing/benchmarking performance and accuracy.

They are not meant for use on large datasets!

Kernels defined here:
- SHOKernel: A stochastic harmonic oscillator (SHO) kernel
- IntegratedSHOKernel: A tinygp/JAX implementation of the integrated SHO kernel from Luhn et al. 2026
"""

import jax
import jax.numpy as jnp
import tinygp

from tinygp.kernels.base import Sum, Product
from tinygp.kernels.quasisep import Sum as qsSum, Product as qsProduct


def extract_leaf_kernels(kernel, all=False):
    """Recursively extract all leaf kernels from a sum or product of kernels"""
    leaf_level = (Sum, qsSum, Product, qsProduct) if all else (Sum, qsSum)
    if isinstance(kernel, leaf_level):
        return extract_leaf_kernels(kernel.kernel1) + extract_leaf_kernels(
            kernel.kernel2
        )
    else:
        return [kernel]


def unpack_coordinates(X1, X2):
    """
    Unpack the input coordinates X1 and X2 into time, instrument ID, and exposure time.

    X1 and X2 can either be
        t1 and t2 for single instrument, no exposure times
    or
        (t1, instid1) and (t2, instid2) for multiple instruments, no exposure times
    or
        (t1, instid1, delta1) and (t2, instid2, delta2) for exposure times, also requires instrument IDs

    Returns
        (t1, instid1, delta1), (t2, instid2, delta2)
        where instid and delta are arrays of zeros if not provided
    """
    if not isinstance(X1, tuple):
        # Single instrument, no exposure times
        t1 = X1
        t2 = X2
        instid1 = delta1 = jnp.zeros_like(t1)
        instid2 = delta2 = jnp.zeros_like(t2)
    elif len(X1) == 2:
        # Multiple instruments, no exposure times
        t1, instid1 = X1
        t2, instid2 = X2
        delta1 = delta2 = jnp.zeros_like(t1)
    elif len(X1) == 3:
        # When using and exposure times
        t1, delta1, instid1 = X1
        t2, delta2, instid2 = X2
    else:
        raise ValueError("X1 and X2 must be tuples of length 1, 2 or 3.")
    return (t1, delta1, instid1), (t2, delta2, instid2)


################# Full (dense) SHO kernel  #################
class SHOKernel(tinygp.kernels.Kernel):
    S: jax.Array | float
    w: jax.Array | float
    Q: jax.Array | float
    rho: jax.Array | float
    tau: jax.Array | float
    sig: jax.Array | float

    def __init__(
        self,
        S=None,
        w=None,
        Q=None,
        rho=None,
        tau=None,
        sig=None,
    ):
        """
        A dense representation of the stochastic harmonic oscillator (SHO)

        Parameters
        ----------
        S : jax.Array | float
            Power at characteristic frequency (m^2/rad/s)
        w : jax.Array | float
            Characteristic frequency (rad/s)
        Q : jax.Array | float
            Quality factor (dimensionless).

        Alternatively, one can give a more physical parameterization
        ----------
            rho : jax.Array | float
                Undamped period of the oscillator (s).
            tau : jax.Array | float
                Damping timescale of the process (s).
            sig : jax.Array | float
                Standard deviation of the process (m/s).
        """

        # Extract parameterization and assign values
        param1 = (S is not None) and (w is not None) and (Q is not None)
        param2 = (rho is not None) and (tau is not None) and (sig is not None)
        if param1:
            self.S = S
            self.w = w
            self.Q = Q
            self.rho = 2 * jnp.pi / self.w
            self.tau = 2 * self.Q / self.w
            self.sig = jnp.sqrt(self.S * self.w * self.Q)
        elif param2:
            self.rho = rho
            self.tau = tau
            self.sig = sig
            self.w = 2 * jnp.pi / self.rho
            self.Q = self.tau * self.w / 2
            self.S = self.sig**2 / (self.w * self.Q)
        else:
            raise ValueError("Must specifiy parameter values!")

    def evaluate(self, X1, X2):
        """
        Calculate the kernel for given pair of times X1 and X2.

        X1 and X2 can either be
            t1 and t2 for single instrument, no exposure times
        or
            (t1, instid1) and (t2, instid2) for multiple instruments, no exposure times
        or
            (t1, delta1, instid1) and (t2, delta2, instid2) for exposure times,
                also requires instrument IDs even if a single instrument
        """

        (t1, delta1, instid1), (t2, delta2, instid2) = unpack_coordinates(X1, X2)

        # Time between pair of observations
        Delta = jnp.abs(t1 - t2)

        # Calculate the kernel
        eta = jnp.sqrt(jnp.abs(1 - 1 / (4 * self.Q**2)))  # damping factor
        k = jnp.cos(eta * self.w * Delta) + 1 / (2 * eta * self.Q) * jnp.sin(
            eta * self.w * Delta
        )
        return self.sig**2 * jnp.exp(-Delta / self.tau) * k

    def __repr__(self):
        s = f"{type(self)}(\n"
        s += f" S={self.S},\n"
        s += f" w={self.w}\n"
        s += f" Q={self.Q}\n"
        s += f" rho={self.rho}\n"
        s += f" tau={self.tau}\n"
        s += f" sig={self.sig}\n"
        s += ")"
        return s


################# Full/dense Integrated SHO kernel  #################
class IntegratedSHOKernel(tinygp.kernels.Kernel):
    S: jax.Array | float
    w: jax.Array | float
    Q: jax.Array | float
    rho: jax.Array | float
    tau: jax.Array | float
    sig: jax.Array | float
    eta: jax.Array | float
    a: jax.Array | float

    def __init__(
        self,
        S=None,
        w=None,
        Q=None,
        rho=None,
        tau=None,
        sig=None,
    ):
        """
        The full/dense representation of the stochastic harmonic oscillator (SHO)
        kernel integrated over a finite time interval

        Details in `Luhn et al. 2026 <https://arxiv.org/abs/2601.02462>`__.

        Parameters
        ----------
        S : jax.Array | float
            Power at characteristic frequency (m^2/rad/s)
        w : jax.Array | float
            Characteristic frequency (rad/s)
        Q : jax.Array | float
            Quality factor (dimensionless).

        Alternatively, one can give a more physical parameterization
        ----------
            rho : jax.Array | float
                Undamped period of the oscillator (s).
            tau : jax.Array | float
                Damping timescale of the process (s).
            sig : jax.Array | float
                Standard deviation of the process (m/s).
        """

        # Extract parameterization and assign values
        param1 = (S is not None) and (w is not None) and (Q is not None)
        param2 = (rho is not None) and (tau is not None) and (sig is not None)
        if param1:
            self.S = S
            self.w = w
            self.Q = Q
            self.rho = 2 * jnp.pi / self.w
            self.tau = 2 * self.Q / self.w
            self.sig = jnp.sqrt(self.S * self.w * self.Q)
        elif param2:
            self.rho = rho
            self.tau = tau
            self.sig = sig
            self.w = 2 * jnp.pi / self.rho
            self.Q = self.tau * self.w / 2
            self.S = self.sig**2 / (self.w * self.Q)
        else:
            raise ValueError("Must specifiy parameter values!")

        self.eta = jnp.sqrt(jnp.abs(1 - 1 / (4 * self.Q**2)))  # damping factor
        self.a = 1 / (2 * self.Q * self.eta)  # common quantity in integrals

    def __repr__(self):
        s = f"{type(self)}(\n"
        s += f" S={self.S},\n"
        s += f" w={self.w}\n"
        s += f" Q={self.Q}\n"
        s += f" rho={self.rho}\n"
        s += f" tau={self.tau}\n"
        s += f" sig={self.sig}\n"
        s += ")"
        return s

    def Delta(self, X1, X2):
        """
        Time difference (absolute value)
        between pairs of observations X1, X2.

        Delta = | t1 - t2 |
        """
        # Extract t1, t2 from X1, X2
        t1 = X1[0] if isinstance(X1, tuple) else X1
        t2 = X2[0] if isinstance(X2, tuple) else X2

        # Time between pair of observations
        Delta = jnp.abs(t1 - t2)
        return Delta

    def latent_process(self, Delta):
        """
        Latent kernel for given time difference Delta
        """
        k = jnp.cos(self.eta * self.w * Delta) + 1 / (2 * self.eta * self.Q) * jnp.sin(
            self.eta * self.w * Delta
        )
        return self.sig**2 * jnp.exp(-Delta / self.tau) * k

    # def latent_process(self, X1, X2):
    #     '''
    #     Latent kernel for a given pair of observations X1 and X2.

    #     X1 and X2 can either be
    #         (t1, texp1) and (t2, texp2) for single instrument
    #     or
    #         (t1, instid1, texp1) and (t2, instid2, texp2) for multiple instruments

    #     t1 and t2 can be a unxt.Quantity with units
    #     '''
    #     return self.latent_process(self.Delta(X1, X2))

    def I0(self, y):
        """Helper function for single integral (Eq. 11 in L26)"""
        return jnp.exp(-self.a * y) * (
            (1 - self.a**2) * jnp.sin(y) - 2 * self.a * jnp.cos(y)
        )

    def I1(self, lower, upper):
        """Helper function for double integrals (Eq. 14 in L26)"""

        def f1(y):
            num = (
                jnp.exp(-self.a * y)
                * (1 - self.a**2)
                * (jnp.cos(y) + self.a * jnp.sin(y))
            )
            den = self.eta * self.w * (1 + self.a**2)
            return num / den

        return f1(upper) - f1(lower)

    def I2(self, lower, upper):
        """Helper function for double integrals (Eq. 15 in L26)"""

        def f2(y):
            num = (
                -2 * self.a * jnp.exp(-self.a * y) * (jnp.sin(y) - self.a * jnp.cos(y))
            )
            den = self.eta * self.w * (1 + self.a**2)
            return num / den

        return f2(upper) - f2(lower)

    def integrated_separate(self, Delta, delta1, delta2):
        """
        The double integral for two non-overlapping observations.
        Depends on the time-lag (Delta) and the two exposure times (delta1, delta2)

        Eq. 17 in Luhn et al. (2026): https://arxiv.org/abs/2601.02462
        """
        # Bounds of integrals
        y1 = self.eta * self.w * ((delta1 + delta2) / 2 + Delta)  # Eq. 12
        y2 = self.eta * self.w * ((delta1 + delta2) / 2 + Delta - delta1)  # Eq. 12
        y3 = self.eta * self.w * ((delta1 - delta2) / 2 + Delta)  # Eq. 12
        y4 = self.eta * self.w * ((delta1 - delta2) / 2 + Delta - delta1)  # Eq. 12

        pre = (self.S * self.Q) / (delta1 * delta2 * self.eta * (1 + self.a**2))
        return pre * (
            self.I1(y1, y2) - self.I2(y1, y2) - self.I1(y3, y4) + self.I2(y3, y4)
        )

    def integrated_overlap(self, delta):
        """
        The double integral for two perfectly overlapping
        observations (i.e. zero time-lag). As such, this
        only depends on the exposure time (delta)

        Eq. 13 in Luhn et al. (2026): https://arxiv.org/abs/2601.02462
        """
        # Bounds of integrals
        y1, y2 = self.eta * self.w * delta, 0.0

        pre = (2 * self.S * self.Q) / (delta * delta * self.eta * (1 + self.a**2))
        return pre * (self.I1(y1, y2) - self.I2(y1, y2) + 2 * self.a * delta)

    def integrated_single(self, Delta, delta):
        """
        The single integral for when one observation
        has zero exposure time (e.g. latent curve)

        Eq. 4 in Luhn et al. (2025, in prep)
        """
        # Bounds of integrals
        y1 = self.eta * self.w * (delta / 2 + Delta)
        y2 = self.eta * self.w * (Delta - delta / 2)

        pre = -(self.S * self.Q) / (delta * self.eta)
        return pre / (1 + self.a**2) * (self.I0(y2) - self.I0(y1))

    def full_single_integral(self, Delta, delta):
        """
        This helper function handles the logic for the single integral case.

        For pairs of observation coordinates separated by Delta, such that
        one is the 0-exposure test point (i.e., the latent curve) and the
        other has exposure time delta, this function handles the logic to
        break up the integral and call self.integrated_single() for the
        sub-exposures where appropriate
        """

        # Define our coordinate system:
        # - Delta must be positive
        # - obs1 starts at t=0
        # - obs1 is the finite exposure, obs2 is instantaneous
        # these are the begin and end times for obs1
        # obs1 spans time p1 to p2
        # obs2 is a single point at p3
        # p1 = 0
        p2 = delta
        p3 = delta / 2 + Delta

        # Initalize output
        result = jnp.zeros_like(Delta)

        # Check if we need to use the instantaneous kernel
        # This will be true if the finite exposure (obs1) is zero and/or
        # significantly less than the SHO timescale to be functionally zero
        timescale = 2 * jnp.pi / self.w
        use_latent = jnp.abs(delta / timescale) < 1e-4
        result = jnp.where(use_latent, self.latent_process(Delta), result)

        ## CASE 1
        # Check if obs2 is outside the exposure of obs1
        # If so, can use the single integral directly
        notothers = ~use_latent
        case1 = p2 <= p3
        result = jnp.where(
            case1 & notothers, self.integrated_single(Delta, delta), result
        )

        ## CASE 2
        # If obs2 did occur during exposure of obs1
        # we need to split the integral into a left and right side
        # result = jnp.where(p3 < p2,
        #         (self.integrated_single(p3/2, p3)*p3
        #         + self.integrated_single(p3/2, delta-p3 )*(delta-p3))/delta, result)
        def calc_integral(p3, delta):
            delta_left = p3
            Delta_left = delta_left / 2
            int = self.integrated_single(Delta_left, delta_left) * delta_left

            delta_right = delta - p3  # or p2-p3
            Delta_right = delta_right / 2
            int += self.integrated_single(Delta_right, delta_right) * delta_right

            return int / delta

        notothers &= ~case1
        case2 = p3 < p2
        result = jnp.where(case2 & notothers, calc_integral(p3, delta), result)
        return result

    def evaluate(self, X1, X2):
        """
        Compute the integrated SHO kernel for a pair of observations X1 and X2.

        X1 and X2 can either be
            (t1, texp1) and (t2, texp2) for single instrument
        or
            (t1, texp1, instid1) and (t2, texp2, instid2) for multiple instruments
        """
        (t1, delta1, instid1), (t2, delta2, instid2) = unpack_coordinates(X1, X2)

        # Time difference and exposure times in units the kernel is defined in
        # Delta = jnp.abs((t1 - t2).to(self.tunit).value)
        # delta1 = delta1.to(self.tunit).value
        # delta2 = delta2.to(self.tunit).value
        Delta = jnp.abs(t1 - t2)

        # Define our coordinate system:
        # - Delta must be positive
        # - obs1 starts at t=0
        # - obs1 is the longer exposure
        delta1, delta2 = jnp.maximum(delta1, delta2), jnp.minimum(delta1, delta2)

        # In our coordinate system, these are the
        # begin and end times for both observations
        # obs1 spans time p1 to p2
        # obs2 spans time p3 to p4
        p1 = 0
        p2 = delta1
        p3 = (delta1 - delta2) / 2 + Delta
        p4 = p3 + delta2

        # Initialize output kernel
        k = jnp.zeros_like(Delta)

        ##### Zero-exposure cases #####
        # If the "longer" exposure (obs1) is zero, or is
        # significantly less than the SHO timescale to be
        # functionally zero, then both obs are zero-exposure
        # so we can simply use the latent kernel with Delta
        timescale = 2 * jnp.pi / self.w
        use_latent = jnp.abs(delta1 / timescale) < 1e-8
        k = jnp.where(use_latent, self.latent_process(Delta), k)

        # We might also have a finite exposure (obs1) but
        # the second is zero or functionally zero.
        # This can come up when conditioning the GP to
        # determine the "true" latent curve
        notothers = ~use_latent
        case0 = jnp.abs(delta2 / timescale) < 1e-8
        k = jnp.where(case0 & notothers, self.full_single_integral(Delta, delta1), k)

        ##### Finite-exposure cases #####

        ##### CASE 1: obs1 and obs2 completely overlap (e.g., the diagonal)
        notothers &= ~case0
        case1 = (Delta == 0) & (delta1 == delta2)
        k = jnp.where(case1 & notothers, self.integrated_overlap(delta1), k)

        ##### CASE 2: obs1 and obs2 are completely separated
        notothers &= ~case1
        case2 = Delta >= (delta1 + delta2) / 2
        k = jnp.where(
            case2 & notothers, self.integrated_separate(Delta, delta1, delta2), k
        )

        ##### CASE 3: obs 1 and obs 2 share mutual partial overlap
        #     |----------|
        #           |-------|
        #     p1    p3   p2 p4
        notothers &= ~case2
        overlap = Delta < (delta1 + delta2) / 2

        case3 = overlap & (p4 > p2)

        def calc_case3(p1, p2, p3, p4, delta1, delta2):
            # Break it up into 3 integrals

            # Int 1 (non-overlap 1)
            # |------|
            #        |------|
            # p1     p3     p4
            # delta1_1 = delta1
            # delta2_1 = (delta1+delta2)/2+Delta - t2
            # Delta_1  = t2+delta2_1/2. - t2/2
            delta1_1 = p3 - p1
            delta2_1 = p4 - p3
            Delta_1 = (p3 + delta2_1 / 2) - (p1 + delta1_1 / 2)
            int_case3 = self.integrated_separate(Delta_1, delta1_1, delta2_1) * (
                delta1_1 * delta2_1
            )
            # Note we multiply by (delta1_1*delta2_1) to unnormalize the integral,
            # so we can renormalize using the full 1/(d1*d2) at the end

            # Int 2 (non-overlap 2)
            #        |----|
            #             |---|
            #        p3   p2  p4
            # delta1_2 = t1
            # delta2_2 = t2-t1
            # Delta_2  = t2-delta2_2/2. - delta1_2/2
            delta1_2 = p2 - p3
            delta2_2 = p4 - p2
            Delta_2 = (p2 + delta2_2 / 2) - (p3 + delta1_2 / 2)
            ## AVOID THIS INTEGRAL IF p2==p3
            int_case3 = jnp.where(
                p2 == p3,
                int_case3,
                int_case3
                + self.integrated_separate(Delta_2, delta1_2, delta2_2)
                * (delta1_2 * delta2_2),
            )

            # Int 3 (overlap)
            #        |----|
            #        |----|
            #        p3   p2
            # delta_overlap = (t2-t1)
            delta_overlap = p2 - p3
            ## AVOID THIS INTEGRAL IF p2==p3
            int_case3 = jnp.where(
                p2 == p3,
                int_case3,
                int_case3 + self.integrated_overlap(delta_overlap) * (delta_overlap**2),
            )

            return int_case3 / (delta1 * delta2)

        k = jnp.where(case3 & notothers, calc_case3(p1, p2, p3, p4, delta1, delta2), k)

        ##### CASE 4: obs 2 is completely within obs 1
        #      |------------|
        #        |-------|
        #     p1 p3     p4 p2
        notothers &= ~case3
        case4 = overlap & (p4 <= p2)  # & (p2!=p4)

        def calc_case4(p1, p2, p3, p4, delta1, delta2):
            # Int 1 (overlap)
            #    |-------|
            #    |-------|
            #    p3      p4
            delta_overlap = delta2
            int_case4 = self.integrated_overlap(delta_overlap) * (delta_overlap**2)

            # Int 2 (non-overlap 1)
            #  |--|
            #     |------|
            #  p1 p3    p4
            delta1_1 = p3 - p1
            delta2_1 = p4 - p3
            Delta_1 = (p3 + delta2_1 / 2) - (p1 + delta1_1 / 2)
            ## AVOID THIS INTEGRAL IF p1==p3
            int_case4 = jnp.where(
                p1 == p3,
                int_case4,
                int_case4
                + self.integrated_separate(Delta_1, delta1_1, delta2_1)
                * (delta1_1 * delta2_1),
            )
            # int_case4 += self.integrated_separate(Delta_1, delta1_1, delta2_1)*(delta1_1*delta2_1)

            # Int 3 (non-overlap 2)
            #          |--|
            #  |-------|
            #  p3     p4  p2
            delta1_2 = p2 - p4
            delta2_2 = p4 - p3
            Delta_2 = (p4 + delta1_2 / 2) - (p3 + delta2_2 / 2)
            ## AVOID THIS INTEGRAL IF p2==p4
            int_case4 = jnp.where(
                p2 == p4,
                int_case4,
                int_case4
                + self.integrated_separate(Delta_2, delta1_2, delta2_2)
                * (delta1_2 * delta2_2),
            )
            # int_case4 += self.integrated_separate(Delta_2, delta1_2, delta2_2)*(delta1_2*delta2_2)

            return int_case4 / (delta1 * delta2)

        k = jnp.where(case4 & notothers, calc_case4(p1, p2, p3, p4, delta1, delta2), k)

        # notothers &= ~case4
        # print( jnp.all( ~notothers  ) ) # check we got all points

        return k



"""
In ``smolgp``, "solvers" provide a swappable low-level interface for
implementing the filtering and smoothing algorithms required to condition
Gaussian Process models. New solvers can be implemented as external packages
or as pull requests to the main ``smolgp`` GitHub project.

The built in solvers are:

1. :class:`StateSpaceSolver`: A solver that uses the standard Kalman filter
   and RTS smoothing algorithms to derive the optimal predictive means and
   covariances for a linear Gaussian state space model. This is the default
   solver, and is intended to be used with the instantaneous kernels
   implemented in `smolgp.kernels.base`. [TODO: compare benchmarking
   to `tinygp.solvers.QuasisepSolver`].

2. :class:`IntegratedStateSpaceSolver`: TODO....

3. :class:`ParallelStateSpaceSolver`: TODO....

4. :class:`ParallelIntegratedStateSpaceSolver`: TODO....

Up to numerical precision, these are all *exact* solvers.

Users generally won't instantiate these solvers directly; ``smolgp`` should
automatically use the appropriate one given the form of the kernel.

TODO: how will the solver handle a mixed kernel where some are integrated
and some are instantaneous? It should default to the IntegratedStateSpaceSolver,
simply skipping updates for the instantaneous parts at the exposure starts,
but this needs to be tested.

The details for the included solvers are given below, but this is a pretty
low-level feature and the details are definitely subject to change!
"""

from smolgp.solvers.solver import StateSpaceSolver as StateSpaceSolver
from smolgp.solvers.parallel.solver import ParallelStateSpaceSolver as ParallelStateSpaceSolver
from smolgp.solvers.integrated.solver import (
    IntegratedStateSpaceSolver as IntegratedStateSpaceSolver,
)
from smolgp.solvers.integrated.parallel.solver import (
    ParallelIntegratedStateSpaceSolver as ParallelIntegratedStateSpaceSolver,
)



from __future__ import annotations

import jax
import jax.numpy as jnp


def ParallelKalmanFilter(kernel, X, y, noise, return_v_S=False):
    """
    Wrapper for the parallel Kalman filter.

    Parameters:
        kernel: StateSpaceModel kernel
        X: data coordinates, e.g. time or (time, texp, instid)
        y: observations
        noise: Noise model

    Returns:
        A:
        b: filtered means
        C: filtered covariances
        eta:
        J:
    """
    H = kernel.observation_model
    Phi = kernel.transition_matrix
    Q = kernel.process_noise
    R = noise.diagonal() if noise is not None else jnp.zeros_like(y)
    m0 = jnp.zeros(kernel.dimension)
    P0 = kernel.stationary_covariance()
    t = kernel.coord_to_sortable(X)

    asso_params = make_associative_params(Phi, H, Q, R, t, y, m0, P0)
    A, b, C, eta, J = parallel_kalman_filter(asso_params)
    m_pred, P_pred, v, S = postprocess(Phi, H, Q, R, X, t, y, b, C, m0, P0)
    # return (A, b, C, eta, J), (m_pred, P_pred, v, S)
    m_filt, P_filt = b, C
    if return_v_S:
        return m_filt, P_filt, m_pred, P_pred, v, S
    else:
        return m_filt, P_filt, m_pred, P_pred


@jax.jit
def make_associative_params(Phi, H, Q, R, t, y, m0, P0):
    """Generate the associative parameters needed for parallel Kalman

    See Eqns. 10, 11, 12 from Sarkka & Garcia-Fernandez (2020)

    """

    def make_first_params(
        Phi,
        H,
        m0,
        P0,
        y0,
        r0,
    ):
        Phi0 = Phi(0, 0)
        H0 = H(0)  # this is sort of unnecessary but we'll keep it for now

        m = Phi0 @ m0
        P = Phi0 @ P0 @ Phi0.T  # Q(0,0) = 0
        S = H0 @ P @ H0.T + r0
        K = jnp.linalg.solve(S.T, (P @ H0.T).T).T

        A = jnp.zeros_like(Phi0)
        b = jnp.squeeze(m + K @ (y0 - H0 @ m))
        C = P - K @ S @ K.T

        _M = Phi0.T @ H0.T
        eta = jnp.squeeze(_M @ jnp.linalg.solve(S, jnp.atleast_1d(y0)))
        J = _M @ jnp.linalg.solve(S, _M.T)

        return (A, b, C, eta, J)

    def make_generic_params(
        Phi,
        H,
        Q,
        t_delta,
        y,
        r,
    ):
        Phi_dt = Phi(0, t_delta)
        I = jnp.eye(Phi_dt.shape[-1])

        Hk = H(t_delta)  # TODO: this is wrong, pass data coordinate here
        Q_dt = Q(0, t_delta)

        S = Hk @ Q_dt @ Hk.T + r
        K = jnp.linalg.solve(S.T, (Q_dt @ Hk.T).T).T

        A = (I - K @ Hk) @ Phi_dt
        b = jnp.squeeze(K @ jnp.atleast_1d(y))  # remove atleast_1d?
        C = (I - K @ Hk) @ Q_dt

        _M = Phi_dt.T @ Hk.T
        eta = jnp.squeeze(_M @ jnp.linalg.solve(S, jnp.atleast_1d(y)))
        J = _M @ jnp.linalg.solve(S, _M.T)

        return (A, b, C, eta, J)

    A0, b0, C0, eta0, J0 = make_first_params(Phi, H, m0, P0, y[0], R[0])
    t_delta = jnp.diff(t)
    A, b, C, eta, J = jax.vmap(
        make_generic_params,
        in_axes=(None, None, None, 0, 0, 0),
    )(Phi, H, Q, t_delta, y[1:], R[1:])

    A_all = jnp.concatenate([A0[jnp.newaxis, ...], A], axis=0)
    b_all = jnp.concatenate([b0[jnp.newaxis, ...], b], axis=0)
    C_all = jnp.concatenate([C0[jnp.newaxis, ...], C], axis=0)
    eta_all = jnp.concatenate([eta0[jnp.newaxis, ...], eta], axis=0)
    J_all = jnp.concatenate([J0[jnp.newaxis, ...], J], axis=0)

    return (A_all, b_all, C_all, eta_all, J_all)


@jax.jit
def _combine_per_pair(left, right):
    """See Eqn. 13 & 14 of Sarkka & Garcia-Fernandez (2020) for
    a the algorithm and notation.

    """
    Ai, bi, Ci, etai, Ji = left
    Aj, bj, Cj, etaj, Jj = right

    dim = Ai.shape[-1]
    I = jnp.eye(dim)

    D = I + Ci @ Jj
    E = I + Jj @ Ci

    Aij = Aj @ jnp.linalg.solve(D, Ai)
    bij = Aj @ jnp.linalg.solve(D, bi + Ci @ etaj) + bj
    Cij = Aj @ jnp.linalg.solve(D, Ci) @ Aj.T + Cj
    etaij = Ai.T @ jnp.linalg.solve(E, etaj - Jj @ bi) + etai
    Jij = Ai.T @ jnp.linalg.solve(E, Jj) @ Ai + Ji

    Cij = 0.5 * (Cij + Cij.T)
    Jij = 0.5 * (Jij + Jij.T)

    return (Aij, bij, Cij, etaij, Jij)


@jax.jit
def parallel_kalman_filter(asso_params):
    """
    Jax implementation of the parallel Kalman filter algorithm

    See Section 4A of Sarkka & Garcia-Fernandez (2020) for
    a detailed description of the algorithm and notation.

    Total runtime (span) complexity is ~O(logN) where N is the number
    of time steps.
    """

    A, b, C, eta, J = jax.lax.associative_scan(
        jax.vmap(_combine_per_pair),
        asso_params,
    )

    return (A, b, C, eta, J)


@jax.jit
def postprocess(Phi, H, Q, R, X, t, y, b, C, m0, P0):
    t_delta = jnp.diff(t)
    dim = b.shape[-1]
    I = jnp.eye(dim)

    Phis = jax.vmap(Phi, in_axes=(None, 0))(0, t_delta)
    Qs = jax.vmap(Q, in_axes=(None, 0))(0, t_delta)

    Phi_all = jnp.concatenate(
        [I[jnp.newaxis, ...], Phis],
        axis=0,
    )
    Q_all = jnp.concatenate(
        [jnp.zeros_like(I)[jnp.newaxis, ...], Qs],
        axis=0,
    )
    H_all = jax.vmap(H, in_axes=(0,))(X)
    R_all = R

    m_prev = jnp.concatenate(
        [m0[jnp.newaxis, ...], b[:-1]],
        axis=0,
    )
    P_prev = jnp.concatenate(
        [P0[jnp.newaxis, ...], C[:-1]],
        axis=0,
    )

    m_pred = jax.vmap(lambda _Phi, _m: _Phi @ _m)(
        Phi_all,
        m_prev,
    )
    P_pred = jax.vmap(lambda _Phi, _P_prev, _Q: _Phi @ _P_prev @ _Phi.T + _Q)(
        Phi_all,
        P_prev,
        Q_all,
    )

    y_pred = jax.vmap(lambda _H, _m: _H @ _m, in_axes=(0, 0))(
        H_all,
        m_pred,
    )

    v = y[..., jnp.newaxis] - y_pred

    S = jax.vmap(lambda _H, _P, _R: _H @ _P @ _H.T + _R)(
        H_all,
        P_pred,
        R_all,
    )

    return (m_pred, P_pred, v, S)



from __future__ import annotations

import jax
import jax.numpy as jnp


def IntegratedKalmanFilter(
    kernel, X, y, t_states, obsid, instid, stateid, noise=None, return_v_S=False
):
    """
    Wrapper for integrated_kalman_filter function

    Parameters:
        kernel  : IntegratedStateSpaceModel kernel
        X       : Array of size N, data coordinates (e.g. (time, texp, instid))
        y       : Array of size N, measurements at the data coordinates
        t_states: Array of size K, sorted time coordinate of all states (exposure starts and ends)
        obsid   : Array of size N, which observation (0,...,N-1) is being made at each state k
        instid  : Array of size N, which instrument (0,...,Ninst-1) recorded observation n
        stateid : Array of size K, 0 for exposure-start, 1 for exposure-end
        noise   : Noise model
        return_v_S : Whether to return innovation and its covariance (for likelihood computation)

    Returns:
        m_filtered : filtered means
        P_filtered : filtered covariances
        m_predicted: predicted means
        P_predicted: predicted covariances
    """

    # Model components
    H_aug = kernel.observation_model
    A_aug = kernel.transition_matrix
    Q_aug = kernel.process_noise
    RESET = kernel.reset_matrix
    R = noise.diagonal() if noise is not None else jnp.zeros_like(y)

    # Initial state and covariance
    # mean = jnp.zeros(kernel.d) # TODO: mean function of base kernel
    # m0 = jnp.block([mean] + kernel.num_insts*[jnp.zeros(kernel.d)])
    m0 = jnp.zeros(kernel.dimension)
    P0 = kernel.stationary_covariance()

    output = integrated_kalman_filter(
        A_aug, Q_aug, H_aug, R, RESET, X, y, t_states, obsid, instid, stateid, m0, P0
    )
    if return_v_S:
        return output
    else:
        m_filtered, P_filtered, m_predicted, P_predicted, v, S = output
        return m_filtered, P_filtered, m_predicted, P_predicted


@jax.jit
def integrated_kalman_filter(
    A_aug, Q_aug, H_aug, R, RESET, X, y, t_states, obsid, instid, stateid, m0, P0
):
    """
    Jax implementation of the integrated Kalman filter algorithm

    See Section 3.2.1 in Rubenzahl & Hattori et al. (in prep)
    for detailed description of the algorithm and notation.
    """

    H = jax.vmap(H_aug)(X)

    @jax.jit
    def step(carry, k):
        # Unpack previous state and covariance
        m_prev, P_prev = carry

        # If k==0 we use the prior m0, Pinf and zero time-lag (dt=0)
        Delta = jax.lax.cond(k > 0, lambda i: t_states[i] - t_states[i - 1], lambda _: 0.0, k)
        n = obsid[k]

        # Get transition matrix
        A_prev = A_aug(0, Delta)
        Q_prev = Q_aug(0, Delta)

        # Predict step is same
        m_pred = A_prev @ m_prev
        P_pred = A_prev @ P_prev @ A_prev.T + Q_prev

        # Update the end of the exposure
        def update_end():
            Hk = H[n]
            y_pred = Hk @ m_pred  # predicted observation
            v_k = y[n] - y_pred  # "innovation" or "surprise" term
            S_k = Hk @ P_pred @ Hk.T + R[n]  # uncertainy in predicted observation
            K_k = jnp.linalg.solve(S_k.T, (P_pred @ Hk.T).T).T  # Kalman gain
            m_k = m_pred + K_k @ v_k  # conditioned state estimate
            P_k = P_pred - K_k @ S_k @ K_k.T  # conditioned covariance estimate
            return m_k, P_k, m_pred, P_pred, v_k, S_k

        # Update the start of the exposure, aka reset its z to zero
        def update_start():
            Reset = RESET(instid[n])
            m_k = Reset @ m_pred
            P_k = Reset @ P_pred @ Reset.T
            Hk = H[n]  # TODO: change this and next two lines to use shapes?
            v_k = jnp.zeros_like(Hk @ m_pred)  # maybe e.g. jax broadcast_shapes?
            S_k = jnp.zeros_like(Hk @ P_pred @ Hk.T)
            return m_k, P_k, m_pred, P_pred, v_k, S_k

        m_k, P_k, m_pred, P_pred, v_k, S_k = jax.lax.cond(
            stateid[k] == 0,
            lambda _: update_start(),
            lambda _: update_end(),
            operand=None,
        )

        return (m_k, P_k), (m_k, P_k, m_pred, P_pred, v_k, S_k)

    # Initialize carry with prior state and covariance
    init_carry = (m0, P0)

    # Run the filter over all time steps, unpack, and return results
    _, outputs = jax.lax.scan(step, init_carry, jnp.arange(len(t_states)))
    m_filtered, P_filtered, m_predicted, P_predicted, v, S = outputs

    # only return v,S at exposure ends (where there is data)
    ends_idx = jnp.nonzero(stateid == 1, size=y.shape[0])[0]
    v_sel = jnp.take(v, ends_idx, axis=0)
    S_sel = jnp.take(S, ends_idx, axis=0)

    return m_filtered, P_filtered, m_predicted, P_predicted, v_sel, S_sel



import tinygp
import smolgp
import jax.numpy as jnp

def unpack_idata(data):
    t_train = data[0, :]
    y_train = data[1, :]
    yerr = data[2, :]
    texp = data[3, :]
    instid = data[4, :].astype(jnp.int64)
    X_train = (t_train, texp, instid)
    return X_train, y_train, yerr

def unpack_data(data):
    t_train = data[0, :]
    y_train = data[1, :]
    yerr = data[2, :]
    return t_train, y_train, yerr

######################################## INSTANTANEOUS DATA FUNCTIONS ########################################
#################### LIKELIHOOD ####################
def ss_llh(data, kernel):
    t_train, y_train, yerr = unpack_data(data)
    gp_ss = smolgp.GaussianProcess(kernel, t_train, diag=yerr**2)
    return gp_ss.log_probability(y_train)

def qs_llh(data, kernel):
    t_train, y_train, yerr = unpack_data(data)
    gp_qs = tinygp.GaussianProcess(kernel, t_train, diag=yerr**2)
    return gp_qs.log_probability(y_train)

def gp_llh(data, kernel):
    t_train, y_train, yerr = unpack_data(data)
    gp_gp = tinygp.GaussianProcess(kernel, t_train, diag=yerr**2)
    return gp_gp.log_probability(y_train)

def pss_llh(data, kernel):
    t_train, y_train, yerr = unpack_data(data)
    gp_ss = smolgp.GaussianProcess(
        kernel, t_train, diag=yerr**2, solver=smolgp.solvers.ParallelStateSpaceSolver
    )
    return gp_ss.log_probability(y_train)

#################### CONDITION ####################
def ss_cond(data, kernel):
    t_train, y_train, yerr = unpack_data(data)
    gp_ss = smolgp.GaussianProcess(kernel, t_train, diag=yerr**2)
    llh, condGP_ss = gp_ss.condition(y_train)
    return jnp.array([condGP_ss.loc, condGP_ss.variance])

def qs_cond(data, kernel):
    t_train, y_train, yerr = unpack_data(data)
    gp_qs = tinygp.GaussianProcess(kernel, t_train, diag=yerr**2)
    llh, condGP_qs = gp_qs.condition(y_train)
    return jnp.array([condGP_qs.loc, condGP_qs.variance])

def gp_cond(data, kernel):
    t_train, y_train, yerr = unpack_data(data)
    gp_gp = tinygp.GaussianProcess(kernel, t_train, diag=yerr**2)
    llh, condGP_gp = gp_gp.condition(y_train)
    return jnp.array([condGP_gp.loc, condGP_gp.variance])

def pss_cond(data, kernel):
    t_train, y_train, yerr = unpack_data(data)
    gp_ss = smolgp.GaussianProcess(
        kernel, t_train, diag=yerr**2, solver=smolgp.solvers.ParallelStateSpaceSolver
    )
    llh, condGP_ss = gp_ss.condition(y_train)
    return jnp.array([condGP_ss.loc, condGP_ss.variance])

#################### PREDICTION ####################
## TODO?: Only time the actual prediction part (use condGP here)
def ss_pred(t_test, gp_ss, y_train):
    mu, var = gp_ss.predict(t_test, y_train, return_var=True)
    return jnp.array([mu, var])

def qs_pred(t_test, gp_qs, y_train):
    mu, var = gp_qs.predict(y_train, t_test, return_var=True)
    return jnp.array([mu, var])

def gp_pred(t_test, gp_gp, y_train):
    mu, var = gp_gp.predict(y_train, t_test, return_var=True)
    return jnp.array([mu, var])



######################################## INTEGRATED DATA FUNCTIONS ########################################
#################### LIKELIHOOD ####################
def iss_llh(data, kernel):
    X_train, y_train, yerr = unpack_idata(data)
    gp_ss = smolgp.GaussianProcess(kernel, X_train, diag=yerr**2)
    return gp_ss.log_probability(y_train)

def igp_llh(data, kernel):
    X_train, y_train, yerr = unpack_idata(data)
    gp_gp = tinygp.GaussianProcess(kernel, X_train, diag=yerr**2)
    return gp_gp.log_probability(y_train)

def ipss_llh(data, kernel):
    X_train, y_train, yerr = unpack_idata(data)
    gp_ss = smolgp.GaussianProcess(
        kernel, X_train, diag=yerr**2, solver=smolgp.solvers.ParallelIntegratedStateSpaceSolver
    )
    return gp_ss.log_probability(y_train)

#################### CONDITION ####################
def iss_cond(data, kernel):
    X_train, y_train, yerr = unpack_idata(data)
    gp_ss = smolgp.GaussianProcess(kernel, X_train, diag=yerr**2)
    llh, condGP_ss = gp_ss.condition(y_train)
    return jnp.array([condGP_ss.loc, condGP_ss.variance])

def igp_cond(data, kernel):
    X_train, y_train, yerr = unpack_idata(data)
    gp_gp = tinygp.GaussianProcess(kernel, X_train, diag=yerr**2)
    llh, condGP_gp = gp_gp.condition(y_train)
    return jnp.array([condGP_gp.loc, condGP_gp.variance])

def ipss_cond(data, kernel):
    X_train, y_train, yerr = unpack_idata(data)
    gp_ss = smolgp.GaussianProcess(
        kernel, X_train, diag=yerr**2, solver=smolgp.solvers.ParallelIntegratedStateSpaceSolver
    )
    llh, condGP_ss = gp_ss.condition(y_train)
    return jnp.array([condGP_ss.loc, condGP_ss.variance])

#################### PREDICTION ####################
def iss_pred(t_test, gp_ss, y_train):
    X_test = (t_test, jnp.zeros_like(t_test), jnp.zeros_like(t_test).astype(int))
    mu, var = gp_ss.predict(X_test, y_train, return_var=True)
    return jnp.array([mu, var])

def igp_pred(t_test, gp_gp, y_train):
    X_test = (t_test, jnp.zeros_like(t_test), jnp.zeros_like(t_test).astype(int))
    mu, var = gp_gp.predict(y_train, X_test, return_var=True)
    return jnp.array([mu, var])



from __future__ import annotations
from typing import Any

import jax
import jax.numpy as jnp
import equinox as eqx

from tinygp.helpers import JAXArray
from tinygp.noise import Noise
from tinygp.solvers.quasisep.solver import QuasisepSolver
from smolgp.kernels.base import StateSpaceModel

from smolgp.solvers.parallel.kalman import ParallelKalmanFilter
from smolgp.solvers.parallel.rts import ParallelRTSSmoother


class ParallelStateSpaceSolver(eqx.Module):
    """
    A solver that uses ``jax.lax.associative_scan`` to implement
    parallel Kalman filtering and RTS smoothing
    """

    X: JAXArray
    kernel: StateSpaceModel
    noise: Noise

    def __init__(
        self,
        kernel: StateSpaceModel,
        X: JAXArray,
        noise: Noise,
    ):
        """Build a :class:`StateSpaceSolver` for a given kernel and coordinates

        Args:
            kernel: The kernel function.
            X: The input coordinates.
            y: The measurement values.
            noise: The noise model for the process.
        """
        self.kernel = kernel
        self.X = X
        self.noise = noise

    def normalization(self) -> JAXArray:
        # TODO: do we want/can we implement this in state space? for now, fall back to quasisep
        return QuasisepSolver(self.kernel, self.X, self.noise).normalization()

    def Kalman(self, y, return_v_S=True) -> Any:
        """Wrapper for Kalman filter used with this solver"""
        return ParallelKalmanFilter(
            self.kernel, self.X, y, self.noise, return_v_S=return_v_S
        )

    def RTS(self, kalman_results) -> Any:
        """Wrapper for RTS smoother used with this solver"""
        return ParallelRTSSmoother(self.kernel, self.X, kalman_results)

    def condition(self, y, return_v_S=False) -> JAXArray:
        """
        Compute the Kalman predicted, filtered, and RTS smoothed
        means and covariances at each of the input coordinates
        """

        # Kalman filtering
        kalman_results = self.Kalman(y, return_v_S=True)
        if return_v_S:
            m_filtered, P_filtered, m_predicted, P_predicted, v, S = kalman_results
            v_S = (v, S)
        else:
            m_filtered, P_filtered, m_predicted, P_predicted = kalman_results
            v_S = None

        # RTS smoothing
        rts_results = self.RTS((m_filtered, P_filtered))
        _, m_smoothed, P_smoothed = rts_results

        # Pack-up results and return
        t_states = self.kernel.coord_to_sortable(self.X)
        conditioned_states = (
            (m_predicted, P_predicted),
            (m_filtered, P_filtered),
            (m_smoothed, P_smoothed),
        )
        return t_states, conditioned_states, v_S

    @jax.jit
    def predict(self, X_test, conditioned_results) -> JAXArray:
        """
        Algorithm for making predictions at arbitrary coordinates X_test

        Args:
            X_test              : The test coordinates.
            conditioned_results : The output of self.condition

        Returns:
            pred_mean : Predicted means of the states at X_test
            pred_var  : Predicted variances of the states at X_test

        There are three cases:
            1. Retrodiction  : smoothing from the first data point
                               using the prior as the prediction
            2. Interpolation : filtering from most recent data point
                               and smoothing from next future point
            3. Extrapolation : predicting from final filtered point
        """

        # Unpack conditioned results
        state_coords, conditioned_states, _ = conditioned_results
        (
            (m_predicted, P_predicted),
            (m_filtered, P_filtered),
            (m_smoothed, P_smoothed),
        ) = conditioned_states
        t_states, instid, obsid, stateid = state_coords

        # Unpack test coordinates
        t_test = self.kernel.coord_to_sortable(X_test)

        N = len(self.X)  # number of data points
        K = len(t_states)  # number of states
        M = len(t_test)  # number of test points

        # Prior covariance for retrodiction
        Pinf = self.kernel.stationary_covariance()
        if not isinstance(Pinf, JAXArray):  # if multicomponent model
            Pinf = Pinf.to_dense()  # needs to be array form here

        # Prior mean for retrodiction
        # mean = jnp.zeros(self.kernel.d)  # TODO: mean function of base kernel
        # m0 = jnp.block([mean] + self.kernel.num_insts * [jnp.zeros(self.kernel.d)])
        m0 = jnp.zeros(self.kernel.dimension)

        # Nearest (future) datapoint
        ## TODO: we assume X_states is sorted here; should we enforce that?
        k_nexts = jnp.searchsorted(t_states, t_test, side="right")

        # Which method to use for each test point:
        past = k_nexts <= 0  # Retrodiction
        future = k_nexts >= K  # Forecast
        during = ~past & ~future  # Interpolate
        cases = past.astype(int) * 0 + during.astype(int) * 1 + future.astype(int) * 2

        # Shorthand for matrices
        A = self.kernel.transition_matrix
        Q = self.kernel.process_noise

        def predict(k, ktest):
            """
            Kalman prediction from most recent
            filtered (not smoothed) state
            """
            dt = X_test[ktest] - self.X[k]
            m_k = m_filtered[k]
            P_k = P_filtered[k]
            A_star = A(0, dt)  # transition matrix from t_k to t_star
            Q_star = Q(0, dt)  # process noise from t_k to t_star
            m_star_pred = A_star @ m_k
            P_star_pred = A_star @ P_k @ A_star.T + Q_star
            # No Kalman update since we have no data at t_star, so we're done
            return m_star_pred, P_star_pred

        def smooth(k_next, ktest, m_star_pred, P_star_pred):
            """
            RTS smooth the prediction (ktest) using
            the nearest future (smoothed) state (k_next)

            m_star_pred and P_star_pred are the output of predict(k, k_star)
            """
            # Next (future) data point predicted & smoothed state
            m_pred_next = m_predicted[
                k_next
            ]  # prediction (no kalman update) at next data point
            P_pred_next = P_predicted[
                k_next
            ]  # prediction (no kalman update) at next data point
            m_hat_next = m_smoothed[k_next]  # RTS smoothed state at next data point
            P_hat_next = P_smoothed[
                k_next
            ]  # RTS smoothed covariance at next data point

            # Time-lag between states
            dt = self.X[k_next] - X_test[ktest]

            # Transition matrix for this step
            A_k = A(0, dt)

            # Compute smoothing gain
            # P_pred_next_inv = jnp.linalg.inv(P_pred_next)
            # G_k = P_star_pred @ A_k.T @ P_pred_next_inv # smoothing gain
            G_k = jnp.linalg.solve(
                P_pred_next.T, (P_star_pred @ A_k.T).T
            ).T  # more stable

            # Update state and covariance
            m_star_hat = m_star_pred + G_k @ (m_hat_next - m_pred_next)
            P_star_hat = P_star_pred + G_k @ (P_hat_next - P_pred_next) @ G_k.T

            return m_star_hat, P_star_hat

        def retrodict(ktest):
            """Reverse-extrapolate from first datapoint t_star"""
            m_star, P_star = smooth(0, ktest, m0, Pinf)
            return m_star, P_star

        def interpolate(ktest):
            """Interpolate between nearest data points"""

            # Get nearest data point before and after the test point
            k_next = k_nexts[ktest]
            k_prev = k_next - 1

            # 1. Kalman predict from most recent data point (in past)
            m_star_pred, P_star_pred = predict(k_prev, ktest)

            # 2. RTS smooth from next nearest data point (in future)
            m_star_hat, P_star_hat = smooth(k_next, ktest, m_star_pred, P_star_pred)

            return m_star_hat, P_star_hat

        def extrapolate(ktest):
            """Kalman predict from from last datapoint t_star"""
            m_star, P_star = predict(-1, ktest)
            return m_star, P_star

        def predict_point(ktest):
            """
            Switch between retrodiction, interpolation, and extrapolation
            for a single test point ktest
            """
            return jax.lax.switch(
                cases[ktest], (retrodict, interpolate, extrapolate), (ktest)
            )

        # Calculate predictions
        ktests = jnp.arange(0, M, 1)
        (pred_mean, pred_var) = jax.vmap(predict_point)(ktests)

        return pred_mean, pred_var



import jax
import jax.numpy as jnp
import tinygp
import smolgp
from tests.utils import allclose, generate_data

jax.config.update("jax_enable_x64", True)

# The tinygp variances include a default noise of machine epsilon
offset = jnp.sqrt(jnp.finfo(jnp.array([0.0])).eps)


def kernel_function(ksmol, ktiny, tol=1e-14, atol=1e-14):
    """
    Check the kernel function k(Delta) is the same
    """
    dts = jnp.linspace(0, 1000, 500)
    zeros = jnp.zeros_like(dts)
    cov_tiny = ksmol(zeros, dts)  # [0, :]
    cov_smol = ktiny(zeros, dts)  # [0, :]
    res = cov_tiny - cov_smol
    allclose("kernel function", res, tol=tol, atol=atol)


def likelihood(gp_smol, gp_tiny, y_train, tol=1e-12, atol=1e-12):
    """
    Check the likelihoods are the same
    """
    llh_smol = gp_smol.log_probability(y_train)
    llh_tiny = gp_tiny.log_probability(y_train)
    res = llh_smol - llh_tiny
    allclose("likelihood", res, tol=tol, atol=atol)


def condition(gp_smol, gp_tiny, y_train, tol=1e-12, atol=1e-12, test_all_cases=False):
    """
    Check the conditioned means/vars are the same
    """
    llh_smol, condGP_smol = gp_smol.condition(y_train)
    llh_tiny, condGP_tiny = gp_tiny.condition(y_train, gp_tiny.X)
    mu_res = condGP_tiny.loc - condGP_smol.loc
    var_res = (condGP_tiny.variance - offset) - condGP_smol.variance
    allclose("conditioned means", mu_res, tol=tol, atol=atol)
    allclose("conditioned variances", var_res, tol=tol, atol=atol)

    if test_all_cases:
        # also check that predict with X_test=None gives the same results
        mu_smol, var_smol = gp_smol.predict(X_test=None, y=y_train, return_var=True)
        res = mu_smol - condGP_smol.loc
        allclose("  predict with X_test=None (mean)", res, tol=tol, atol=atol)
        res = var_smol - condGP_smol.variance
        allclose("  predict with X_test=None (var)", res, tol=tol, atol=atol)

        # check full state returns correctly
        m_smol, P_smol = gp_smol.predict(
            X_test=None, y=y_train, return_var=True, return_full_state=True
        )
        m = condGP_smol.states.smoothed_mean
        P = condGP_smol.states.smoothed_cov
        res = m_smol - m
        allclose("  full state predict (mean)", res, tol=tol, atol=atol)
        res = P_smol - P
        allclose("  full state predict (var)", res, tol=tol, atol=atol)


def predict(gp_smol, gp_tiny, y_train, tol=1e-12, atol=1e-12):
    """
    Check the predicted means/vars are the same
    """
    t_train = gp_smol.kernel.coord_to_sortable(gp_smol.X)
    tmin, tmax = t_train.min(), t_train.max()
    dt = 0.1 * (tmax - tmin)  # include a retrodict/extrapolate
    t_test = jnp.linspace(tmin - dt, tmax + dt, 1000)
    if isinstance(gp_smol.X, tuple):
        zeros = jnp.zeros_like(t_test)
        # no exposure time, single instrument
        X_test = (t_test, zeros, zeros.astype(int))
    else:
        X_test = t_test
    mu_tiny, var_tiny = gp_tiny.predict(y_train, X_test, return_var=True)
    mu_smol, var_smol = gp_smol.predict(X_test, y_train, return_var=True)
    mu_res = mu_tiny - mu_smol
    var_res = (var_tiny - offset) - var_smol
    allclose("predicted means", mu_res, tol=tol, atol=atol)
    allclose("predicted variances", var_res, tol=tol, atol=atol)


def kernel(kernel_smol, kernel_tiny, test_all_cases=False):
    """
    Check the smolgp kernel produces the same
    results as tinygp for all the above tests
    """
    if kernel_smol.name in ["ExpSineSquared"]:
        covtol, covatol = 1e-3, 1e-6
        llhtol, llhatol = 0.2, 1e-3
        condtol, condatol = 1e-1, 1e-3
        predtol, predatol = 1e-1, 1e-3
    else:
        covtol, covatol = 1e-9, 1e-14
        llhtol, llhatol = 1e-9, 1e-12
        condtol, condatol = 1e-9, 1e-13
        predtol, predatol = 1e-9, 1e-13

    # Test k(Delta) agrees
    kernel_function(kernel_smol, kernel_tiny, tol=covtol, atol=covatol)

    ## Generate mock data
    N = 50
    yerr = 0.3
    t_train, y_train = generate_data(N, kernel_tiny, yerr, tmin=0, tmax=1000)
    yerr_train = jnp.full_like(t_train, yerr)

    # Build GP objects
    gp_smol = smolgp.GaussianProcess(kernel=kernel_smol, X=t_train, diag=yerr_train**2)
    gp_tiny = tinygp.GaussianProcess(kernel=kernel_tiny, X=t_train, diag=yerr_train**2)

    # Check likelihood
    likelihood(gp_smol, gp_tiny, y_train, tol=llhtol, atol=llhatol)

    # Check conditioning
    condition(
        gp_smol,
        gp_tiny,
        y_train,
        tol=condtol,
        atol=condatol,
        test_all_cases=test_all_cases,
    )

    # Check predictions
    predict(gp_smol, gp_tiny, y_train, tol=predtol, atol=predatol)


def test_kernels():
    kernels = {}

    # Kernel parameters to use for testing
    sigma = 2.1  # amplitude
    scale = 83.3  # length scale for Exp/Matern
    omega = 2 * jnp.pi / 83.3  # freq for SHO/periodic

    ################ Exp ################
    ksmol = smolgp.kernels.Exp(scale=scale, sigma=sigma)
    ktiny = tinygp.kernels.quasisep.Exp(scale=scale, sigma=sigma)
    kernels["Exp"] = [ksmol, ktiny]

    ################ Matern-3/2 ################
    ksmol = smolgp.kernels.Matern32(scale=scale, sigma=sigma)
    ktiny = tinygp.kernels.quasisep.Matern32(scale=scale, sigma=sigma)
    kernels["Matern32"] = [ksmol, ktiny]

    ################ Matern-5/2 ################
    ksmol = smolgp.kernels.Matern52(scale=scale, sigma=sigma)
    ktiny = tinygp.kernels.quasisep.Matern52(scale=scale, sigma=sigma)
    kernels["Matern52"] = [ksmol, ktiny]

    ################ Cosine ################
    ksmol = smolgp.kernels.Cosine(scale=scale, sigma=sigma)
    ktiny = tinygp.kernels.quasisep.Cosine(scale=scale, sigma=sigma)
    kernels["Cosine"] = [ksmol, ktiny]

    ################ SHO ################
    quality = {"underdamped": 5.3, "critical": 0.5, "overdamped": 0.1}
    for shotype in quality:
        q = quality[shotype]
        ksmol = smolgp.kernels.SHO(omega, q, sigma)
        ktiny = tinygp.kernels.quasisep.SHO(omega, q, sigma)
        kernels[f"SHO_{shotype}"] = [ksmol, ktiny]

    ################ ExpSineSquared ################
    ell = 1.164  # 0.7
    gamma = 2 / ell**2
    period = 163.3
    # we have to hardcode sigma=1 to match the definition in tinygp
    ksmol = smolgp.kernels.ExpSineSquared(gamma=gamma, period=period, sigma=1.0)
    ktiny = tinygp.kernels.ExpSineSquared(gamma=gamma, scale=period)
    kernels["ExpSineSquared"] = [ksmol, ktiny]

    ## Test them
    for name in kernels:
        ksmol, ktiny = kernels[name]
        print(f"Testing {name}...")
        kernel(ksmol, ktiny, test_all_cases=name == "Exp")
        print()


if __name__ == "__main__":
    test_kernels()
    print("All kernel tests passed.")



from __future__ import annotations

import jax
import jax.numpy as jnp


def ParallelIntegratedKalmanFilter(
    kernel,
    X,
    y,
    t_states,
    obsid,
    instid,
    stateid,
    noise=None,
    return_v_S=False,
):
    """
    Wrapper for parallel_integrated_kalman_filter function

    Parameters:
        kernel  : IntegratedStateSpaceModel kernel
        X       : Array of size N, data coordinates (e.g. (time, texp, instid))
        y       : Array of size N, measurements at the data coordinates
        t_states: Array of size K, sorted time coordinate of all states (exposure starts and ends)
        obsid   : Array of size N, which observation (0,...,N-1) is being made at each state k
        instid  : Array of size N, which instrument (0,...,Ninst-1) recorded observation n
        stateid : Array of size K, 0 for exposure-start, 1 for exposure-end
        noise   : Noise model
        return_v_S : Whether to return innovation and its covariance (for likelihood computation)

    Returns:
        m_filtered : filtered means
        P_filtered : filtered covariances
        m_predicted: predicted means
        P_predicted: predicted covariances
    """

    H_aug = kernel.observation_model
    Phi_aug = kernel.transition_matrix
    Q_aug = kernel.process_noise
    RESET = kernel.reset_matrix
    R = noise.diagonal() if noise is not None else jnp.zeros_like(y)

    m0 = jnp.zeros(kernel.dimension)
    P0 = kernel.stationary_covariance()

    asso_params = make_associative_params(
        Phi_aug,
        H_aug,
        Q_aug,
        RESET,
        R,
        X,
        y,
        t_states,
        obsid,
        instid,
        stateid,
        m0,
        P0,
    )
    A, b, C, eta, J = parallel_integrated_kalman_filter(asso_params)
    m_pred, P_pred, v, S = postprocess(
        Phi_aug,
        Q_aug,
        H_aug,
        R,
        X,
        y,
        t_states,
        obsid,
        stateid,
        b,
        C,
        m0,
        P0,
    )
    m_filt, P_filt = (b, C)
    if return_v_S:
        return m_filt, P_filt, m_pred, P_pred, v, S
    else:
        return m_filt, P_filt, m_pred, P_pred


@jax.jit
def make_associative_params(
    Phi_aug,
    H_aug,
    Q_aug,
    RESET,
    R,
    X,
    y,
    t_states,
    obsid,
    instid,
    stateid,
    m0,
    P0,
):
    """
    Generate the associative parameters needed for parallel Kalman.
    See Eqns. 10, 11, 12 from Sarkka & Garcia-Fernandez (2020)
    """

    # precompute H at data coordinates
    H_array = jax.vmap(H_aug)(X)  # index with obsid

    state_dim = H_array[0].shape[-1]

    def make_first_params(
        Phi_aug,
        m0,
        P0,
        Reset,
    ):
        Phi0 = Phi_aug(0, 0)

        transition = Reset @ Phi0

        m = transition @ m0
        P = transition @ P0 @ transition.T  # Q(0,0) = 0

        A = Reset
        b = jnp.squeeze(m)
        C = P

        eta = jnp.zeros(state_dim)
        J = jnp.zeros_like(Phi0)

        return (A, b, C, eta, J)

    def to_start_params(ops):
        (
            Phi_dt,
            Q_dt,
            Reset,
            obsid,
        ) = ops

        A = Reset @ Phi_dt
        b = jnp.zeros(Phi_dt.shape[-1])
        C = Reset @ Q_dt @ Reset.T
        eta = jnp.zeros(state_dim)
        J = jnp.zeros_like(Phi_dt)
        return (A, b, C, eta, J)

    def to_end_params(ops):
        (
            Phi_dt,
            Q_dt,
            Reset,
            obsid,
        ) = ops

        I_nx = jnp.eye(Phi_dt.shape[-1])

        Hk = H_array[obsid]
        yk = y[obsid]
        rk = R[obsid]

        Sk = Hk @ Q_dt @ Hk.T + rk
        Kk = jnp.linalg.solve(Sk.T, (Q_dt @ Hk.T).T).T
        factor = I_nx - Kk @ Hk
        A = factor @ Phi_dt
        b = jnp.squeeze(Kk @ jnp.atleast_1d(yk))
        C = factor @ Q_dt
        C = 0.5 * (C + C.T)

        _a = Phi_dt.T @ Hk.T
        _b = jnp.linalg.solve(Sk, jnp.atleast_1d(yk))
        eta = _a @ _b
        _c = jnp.linalg.solve(Sk, Hk @ Phi_dt)
        J = _a @ _c
        return (A, b, C, eta, J)

    A0, b0, C0, eta0, J0 = make_first_params(
        Phi_aug,
        m0,
        P0,
        RESET(instid[0]),
    )
    t_delta = jnp.diff(t_states)

    Phis = jax.vmap(
        Phi_aug,
        in_axes=(
            None,
            0,
        ),
    )(0, t_delta)

    Qs = jax.vmap(
        Q_aug,
        in_axes=(
            None,
            0,
        ),
    )(0, t_delta)

    Resets = jax.vmap(RESET)(instid[1:])
    ops = (
        Phis,
        Qs,
        Resets,
        obsid[1:],
    )
    A, b, C, eta, J = jax.vmap(
        lambda sid, op: jax.lax.cond(
            sid == 0,
            to_start_params,
            to_end_params,
            op,
        ),
    )(stateid[1:], ops)
    A_all = jnp.concatenate([A0[jnp.newaxis, ...], A], axis=0)
    b_all = jnp.concatenate([b0[jnp.newaxis, ...], b], axis=0)
    C_all = jnp.concatenate([C0[jnp.newaxis, ...], C], axis=0)
    eta_all = jnp.concatenate([eta0[jnp.newaxis, ...], eta], axis=0)
    J_all = jnp.concatenate([J0[jnp.newaxis, ...], J], axis=0)

    return (A_all, b_all, C_all, eta_all, J_all)


def _combine_per_pair(left, right):
    """
    See Eqn. 13 & 14 of Sarkka & Garcia-Fernandez (2020) for
    a the algorithm and notation.
    """
    Ai, bi, Ci, etai, Ji = left
    Aj, bj, Cj, etaj, Jj = right

    dim = Ai.shape[-1]
    I = jnp.eye(dim)

    D = I + Ci @ Jj
    E = I + Jj @ Ci

    Aij = Aj @ jnp.linalg.solve(D, Ai)
    bij = Aj @ jnp.linalg.solve(D, bi + Ci @ etaj) + bj
    Cij = Aj @ jnp.linalg.solve(D, Ci) @ Aj.T + Cj
    etaij = Ai.T @ jnp.linalg.solve(E, etaj - Jj @ bi) + etai
    Jij = Ai.T @ jnp.linalg.solve(E, Jj) @ Ai + Ji

    Cij = 0.5 * (Cij + Cij.T)
    Jij = 0.5 * (Jij + Jij.T)

    return (Aij, bij, Cij, etaij, Jij)


@jax.jit
def parallel_integrated_kalman_filter(asso_params):
    """
    Jax implementation of the parallel Kalman filter algorithm
    for integrated measurements.

    See Section 4A of Sarkka & Garcia-Fernandez (2020) for
    a detailed description of the algorithm and notation,
    and section 3.2.4 of Rubenzahl & Hattori et al. (2025)
    for the integrated measurement case.

    Total runtime (span) complexity is O(N/T + logT) where N is the
    number of time steps and T is the number of parallel threads.
    """

    A, b, C, eta, J = jax.lax.associative_scan(
        jax.vmap(_combine_per_pair),
        asso_params,
    )

    return (A, b, C, eta, J)


@jax.jit
def _calc_kf_predictions(
    Phi_aug,
    Q_aug,
    H_aug,
    t_states,
    b,
    C,
    m0,
    P0,
):
    t_delta = jnp.diff(t_states)
    Phi0 = Phi_aug(0, 0)
    Phik = jax.vmap(Phi_aug, in_axes=(None, 0))(0, t_delta)
    Phis = jnp.concatenate([Phi0[jnp.newaxis, ...], Phik], axis=0)

    Q0 = Q_aug(0, 0)
    Qk = jax.vmap(Q_aug, in_axes=(None, 0))(0, t_delta)
    Qs = jnp.concatenate([Q0[jnp.newaxis, ...], Qk], axis=0)

    m_prev = jnp.concatenate(
        [m0[jnp.newaxis, ...], b[:-1]],
        axis=0,
    )
    P_prev = jnp.concatenate(
        [P0[jnp.newaxis, ...], C[:-1]],
        axis=0,
    )

    m_pred = jax.vmap(lambda _Phi, _m: _Phi @ _m)(Phis, m_prev)
    P_pred = jax.vmap(lambda _Phi, _P_prev, _Q: _Phi @ _P_prev @ _Phi.T + _Q)(
        Phis,
        P_prev,
        Qs,
    )

    return (m_pred, P_pred)


@jax.jit
def _calc_vS(
    H_aug,
    R,
    m_pred,
    P_pred,
    X,
    y,
    stateid,
    obsid,
):
    ends_idx = jnp.nonzero(
        stateid == 1,
        size=y.shape[0],
    )[0]  # end states where there is a measurement
    obsidx_in_ends_order = jnp.take(obsid, ends_idx)

    mm = jnp.take(m_pred, ends_idx, axis=0)
    Pm = jnp.take(P_pred, ends_idx, axis=0)
    Xm = jax.tree.map(lambda x: jnp.take(x, obsidx_in_ends_order), X)
    Hm = jax.vmap(H_aug)(Xm)
    ym = jnp.take(y, obsidx_in_ends_order)
    Rm = jnp.take(R, obsidx_in_ends_order)

    y_pred = jax.vmap(
        lambda H, m: H @ m,
        in_axes=(0, 0),
    )(Hm, mm)

    v = ym[..., jnp.newaxis] - y_pred
    S = jax.vmap(
        lambda H, P, R: H @ P @ H.T + R,
        in_axes=(0, 0, 0),
    )(
        Hm,
        Pm,
        Rm,
    )

    return (v, S)


@jax.jit
def postprocess(
    Phi_aug,
    Q_aug,
    H_aug,
    R,
    X,
    y,
    t_states,
    obsid,
    stateid,
    b,
    C,
    m0,
    P0,
):
    m_pred, P_pred = _calc_kf_predictions(
        Phi_aug,
        Q_aug,
        H_aug,
        t_states,
        b,
        C,
        m0,
        P0,
    )

    v, S = _calc_vS(
        H_aug,
        R,
        m_pred,
        P_pred,
        X,
        y,
        stateid,
        obsid,
    )

    return (m_pred, P_pred, v, S)



from __future__ import annotations

import jax
import jax.numpy as jnp
from tinygp.helpers import JAXArray


def KalmanFilter(kernel, X, y, noise, return_v_S=False):
    """
    Wrapper for jitted kalman_filter function

    Parameters:
        kernel: StateSpaceModel kernel
        X: data coordinates, e.g. time or (time, texp, instid)
        y: observations at data coordinates
        noise: Noise model

    Returns:
        m_filtered: filtered means
        P_filtered: filtered covariances
        m_predicted: predicted means
        P_predicted: predicted covariances
    """
    H = kernel.observation_model
    A = kernel.transition_matrix
    Q = kernel.process_noise
    R = noise.diagonal() if noise is not None else jnp.zeros_like(y)
    m0 = jnp.zeros(kernel.dimension)
    P0 = kernel.stationary_covariance()
    if not isinstance(P0, JAXArray):
        P0 = P0.to_dense()  # needed for carry in jax.lax.scan

    t = kernel.coord_to_sortable(X)
    output = kalman_filter(A, Q, H, R, t, y, m0, P0)
    if return_v_S:
        return output
    else:
        m_filtered, P_filtered, m_predicted, P_predicted, v, S = output
        return m_filtered, P_filtered, m_predicted, P_predicted


@jax.jit
def kalman_filter(A, Q, H, R, t, y, m0, P0):
    """
    Jax implementation of the Kalman filter algorithm

    See Theorem 4.2 (pdf page 77) in "Bayesian Filtering and Smoothing"
    by Simo S{\"a}rkk{\"a} for detailed description of the algorithm and notation.

    e.g. _prev is _{k-1}
         _pred is _k^{-}

    Total runtime complexity is O(N*d^3) where N is the number
    of time steps and d is the dimension of the state vector.
    """
    N = len(t)  # number of data points

    def step(carry, k):
        """
        Routine for a single step of the Kalman filter

        Parameters:
            carry: (x_prev, P_prev) - previous state and covariance
            k: index of the current time step

        Returns:
        - Conditioned state (m_k) and covariance (P_k) to carry to next iteration
        - Full output for completed scan (m_k, P_k, m_pred, P_pred)
        """

        # Unpack previous state and covariance
        m_prev, P_prev = carry

        # Logic to check if first time step:
        # If k==0 we use the prior x0, P0
        # and zero time-lag (Delta=0)
        Delta = jax.lax.cond(
            k > 0,
            lambda i: t[i] - t[i - 1],
            lambda _: 0.0,
            k,
        )

        # Get transition matrix
        A_prev = A(0, Delta)
        Q_prev = Q(0, Delta)

        # Predict (Eq. 4.20)
        m_pred = A_prev @ m_prev
        P_pred = A_prev @ P_prev @ A_prev.T + Q_prev

        # Update (Eq. 4.21)
        ## TODO: let t be a tuple and use coord_to_sortable to get time axis out
        ##       That way we can pass the full t into H and it can use e.g. instid etc.
        H_k = H(t[k])  # observation model for this time step
        y_pred = H_k @ m_pred  # predicted observation
        v_k = y[k] - y_pred  # "innovation" or "surprise" term
        S_k = H_k @ P_pred @ H_k.T + R[k]  # uncertainy in predicted observation
        # S_k_inv = jnp.linalg.inv(S_k)
        # K_k = P_pred @ H_k.T @ S_k_inv    # Kalman gain
        K_k = jnp.linalg.solve(S_k.T, (P_pred @ H_k.T).T).T  # more stable
        m_k = m_pred + K_k @ v_k  # conditioned state estimate
        P_k = P_pred - K_k @ S_k @ K_k.T  # conditioned covariance estimate

        return (m_k, P_k), (m_k, P_k, m_pred, P_pred, v_k, S_k)

    # Initialize carry with prior state and covariance
    init_carry = (m0, P0)

    # Run the filter over all time steps, unpack, and return results
    _, outputs = jax.lax.scan(step, init_carry, jnp.arange(N))
    return outputs



from abc import abstractmethod
import time
import numpy as np
import pickle
import os
import jax
import jax.numpy as jnp
import tinygp
import smolgp
from funcs import unpack_data, unpack_idata
from .. import utils

import re
import psutil
import threading
import subprocess
import multiprocessing as mp

mp.set_start_method("spawn", force=True)

key = jax.random.PRNGKey(0)


def format_bytes(n):
    if n == 0:
        return "0 B"
    units = ["B", "KB", "MB", "GB", "TB", "PB"]
    size = float(n)
    idx = 0
    while size >= 1024 and idx < len(units) - 1:
        size /= 1024
        idx += 1
    return f"{size:.1f} {units[idx]}"


def get_data(true_kernel, N, yerr=0.3, exposure_quantities=None, save=True):
    # Generate data of length N
    if exposure_quantities:
        texp, readout = exposure_quantities
        t_train, y_train = utils.generate_integrated_data(
            N, true_kernel, texp=texp, readout=readout, yerr=yerr
        )
        texp_train = jnp.full_like(t_train, texp)
        yerr_train = jnp.full_like(t_train, yerr)
        instid = jnp.full_like(t_train, 0)
        data = jnp.array([t_train, y_train, yerr_train, texp_train, instid])
        savename = f"data/{N}_int.npz"
    else:
        t_train, y_train = utils.generate_data(N, true_kernel, yerr=yerr)
        yerr_train = jnp.full_like(t_train, yerr)
        data = jnp.array([t_train, y_train, yerr_train])
        savename = f"data/{N}.npz"
    if save:
        jnp.savez(savename, data)
    return data


def save_benchmark_data(filename, Ns, runtime, memory, outputs):
    import pickle

    data = {
        "Ns": Ns,
        "runtime": runtime,
        "memory": memory,
        "outputs": outputs,
    }
    with open(filename, "wb") as f:
        pickle.dump(data, f)


def load_benchmark_data(filename, unpack=False):
    import pickle

    with open(filename, "rb") as f:
        data = pickle.load(f)
    if unpack:
        return data["Ns"], data["runtime"], data["memory"], data["outputs"]
    else:
        return data


class MemorySampler:
    def __init__(self, interval=1e-3):
        self.interval = interval
        self.running = False
        self.peak = 0
        self.baseline = 0
        self.proc = psutil.Process(os.getpid())

    @abstractmethod
    def fetch_memory(self):
        raise NotImplementedError

    def _sample(self):
        while self.running:
            mem = self.fetch_memory()
            self.peak = max(self.peak, mem)
            time.sleep(self.interval)

    def start(self):
        self.running = True
        t = threading.Thread(target=self._sample)
        t.daemon = True
        t.start()

    def record_baseline(self, interval=0.1):
        mem = []
        for _ in range(int(interval / self.interval)):
            mem.append(self.fetch_memory())
            time.sleep(self.interval)
        self.baseline = np.mean(mem)

    def stop(self):
        self.running = False


class CPUMemorySampler(MemorySampler):
    def fetch_memory(self):
        return self.proc.memory_info().rss


def get_gpu_processes():
    """
    Parse the 'Processes' section of NVIDIA SMI
    Returns dict {pid: used_bytes}.
    """

    pid_re = re.compile(r"Process ID\s*:\s*(\d+)")
    mem_re = re.compile(r"Used GPU Memory\s*:\s*([\d\.]+)\s*MiB", re.IGNORECASE)
    name_re = re.compile(r"Name\s*:\s*(.*)")
    type_re = re.compile(r"Type\s*:\s*(\S)")

    # Get the "Processes" section of nvidia-smi
    cmd = "nvidia-smi --query"
    out = subprocess.check_output(cmd, shell=True).decode()
    lines = out.split("\n")
    for i, line in enumerate(lines):
        if line.lstrip().startswith("Processes"):
            break

    results = {}
    for line in lines[i:]:
        line = line.strip()

        # Process ID
        m_pid = pid_re.search(line)
        if m_pid:
            # Start a new record
            pid = int(m_pid.group(1))
            results[pid] = {}
            continue

        # Memory
        m_mem = mem_re.search(line)
        if m_mem:
            mem_mib = float(m_mem.group(1))
            mem = int(mem_mib * 1024 * 1024)
            results[pid]["used_memory"] = mem
            continue

        for reg in [name_re, type_re]:
            m = reg.search(line)
            if m:
                name = reg.pattern.split("\\")[0].lower()
                results[pid][name] = m.group(1)
                continue

    return results


class GPUMemorySampler(MemorySampler):
    def fetch_memory(self):
        smi = get_gpu_processes()
        tot_mem = 0
        for pid in smi:
            # if 'smolgp' in smi[pid]['name']:
            #     tot_mem += smi[pid]['used_memory']
            if smi[pid]["type"] == "C":
                tot_mem += smi[pid]["used_memory"]
        return tot_mem

    def record_baseline(self, interval=0.1):
        return 0


def tracer(fn_bytes, dat_bytes, obj_bytes, args_bytes, return_pipe, machine):
    """
    Time and trace memory of a JAX function inside an isolated subprocess.
    """
    # Unpickle function and arguments inside isolated subprocess
    fn = pickle.loads(fn_bytes)
    dat = pickle.loads(dat_bytes)  # this should be JAXArray
    obj = pickle.loads(obj_bytes)  # this can be anything (e.g. kernel/gp)
    args = pickle.loads(args_bytes)  # extra args for fn

    # Create the jitted function here
    @jax.jit
    def fn_jit(x):
        return fn(x, obj, *args)

    # Warm up (JIT compilation)
    out = fn_jit(dat)
    if hasattr(out, "block_until_ready"):
        out.block_until_ready()

    # Benchmarking/tracing
    peak_mem = 0
    interval = 1e-2 if machine == "gpu" else 1e-5
    while peak_mem == 0:
        if machine == "gpu":
            sampler = GPUMemorySampler(interval=interval)
        elif machine == "cpu":
            sampler = CPUMemorySampler(interval=interval)
        else:
            raise ValueError(f"Unknown machine type: {machine}")

        sampler.start()
        sampler.record_baseline(0.1)
        # Time the function with JAX block_until_ready
        start = time.perf_counter()
        out = fn_jit(dat)
        if hasattr(out, "block_until_ready"):
            out.block_until_ready()
        end = time.perf_counter()
        time.sleep(0.1)
        sampler.stop()
        peak_mem = sampler.peak - sampler.baseline
        # Repeat if no memory usage detected
        # (function ran faster than sampler.interval)
        # try a faster interval
        interval /= 10

    # Return both result (pickled) and stats
    return_pipe.send(
        {
            "output": pickle.dumps(out),
            "runtime": end - start,
            "peak_mem": peak_mem,
        }
    )
    return_pipe.close()


def profile_jax_function(
    fn, data, obj, *args, n_repeat=5, machine="cpu", drop_outliers=False, **kwargs
):
    """
    JAX profiler for time benchmarking and memory tracing a function.
    """
    fn_bytes = pickle.dumps(fn)
    dat_bytes = pickle.dumps(data)
    obj_bytes = pickle.dumps(obj)
    args_bytes = pickle.dumps(args)

    runtimes = []
    peaks = []
    output = None

    for _ in range(n_repeat):
        parent_conn, child_conn = mp.Pipe()
        p = mp.Process(
            target=tracer,
            args=(fn_bytes, dat_bytes, obj_bytes, args_bytes, child_conn, machine),
        )
        p.start()
        result = parent_conn.recv()
        p.join()

        runtimes.append(result["runtime"])
        peaks.append(result["peak_mem"])
        output = pickle.loads(result["output"])
    runtimes = np.array(runtimes)
    peaks = np.array(peaks)
    if drop_outliers:
        runtimes = np.delete(runtimes, [runtimes.argmin(), runtimes.argmax()])
        peaks = np.delete(peaks, [peaks.argmin(), peaks.argmax()])

    return (
        (np.mean(runtimes), np.std(runtimes)),
        (np.mean(peaks), np.std(peaks)),
        output,
    )


def benchmark(
    funcs,
    data,
    objs,
    *args,
    n_repeat=5,
    cutoffs={},
    drop_outliers=False,
    use_gpu_profiler=False,
):
    """
    Given some (to-be-jitted) functions, benchmark their runtimes over a range of input sizes.

    Parameters
    ----------
    funcs : list of callables
        List of functions to benchmark. Each function should take a single input array.
    data : list of tuples
        List of data tuples (t_train, y_train, yerr) for each input size

    Returns
    -------
    Ns : list of input sizes
    runtime : dict
        Dictionary mapping function names to lists of runtimes (means and stds) for each input size.
    memory : dict
        Dictionary mapping function names to lists of memory usages (means and stds) for each input size.
    outputs : dict
        Dictionary mapping function names to lists of outputs for each input size.
    """

    runtime = {name: [] for name in funcs}
    memory = {name: [] for name in funcs}
    outputs = {name: [] for name in funcs}
    Ns = []
    machine = "gpu" if use_gpu_profiler else "cpu"
    for n in range(len(data)):
        N = data[n].shape[-1]
        Ns.append(N)
        print(f"  ({n + 1}/{len(data)}):  N = {N}")
        for name in funcs:
            func = funcs[name]
            obj = objs[name]  # either kernel or gp
            cutoff = cutoffs.get(name, 3e4)

            if N <= cutoff:
                t, mem, val = profile_jax_function(
                    func,
                    data[n],
                    obj,
                    *args,
                    n_repeat=n_repeat,
                    machine=machine,
                    drop_outliers=drop_outliers,
                )
                basestr = f"    {name}: time = {t[0]:.4f} ± {t[1]:.4f} s"
                memstr = f", mem = {format_bytes(mem[0])} ± {format_bytes(mem[1])}"
                print(basestr + memstr)
            else:
                t, mem, val = (jnp.nan, jnp.nan), (jnp.nan, jnp.nan), jnp.nan
                print(f"    {name}: Skipped (N={N} > cutoff={cutoff})")

            runtime[name].append(t)
            memory[name].append(mem)
            outputs[name].append(val)

            save_benchmark_data(
                f"results/individual/{func.__name__}_{N}_{machine}.pkl",
                [N],
                {name: [t]},
                {name: [mem]},
                {name: [val]},
            )

    return Ns, runtime, memory, outputs


def run_benchmark(
    true_kernel,
    funcs,
    kernels,
    yerr=0.3,
    N_N=10,
    logN_min=1,
    logN_max=7,
    n_repeat=5,
    cutoffs={},
    drop_outliers=False,
    use_gpu_profiler=False,
    exposure_quantities=None,
):
    """
    Generate data and benchmark the provided functions over a range of input sizes.
    """
    print("Generating data for benchmarking...")
    ## Generate all data ahead of time
    isint = "_int" if exposure_quantities else ""
    Ns = jnp.logspace(logN_min, logN_max, N_N).astype(int)
    data = []
    for N in Ns:
        datafile = f"data/{N}{isint}.npz"
        if os.path.exists(datafile):
            d = jnp.load(datafile)["arr_0"]
            print("  Loaded data from", datafile)
        else:
            print("  Generating data for N =", N)
            d = get_data(
                true_kernel,
                N,
                yerr=yerr,
                exposure_quantities=exposure_quantities,
                save=True,
            )
            print("  Generated and saved data to", datafile)
        data.append(d)

    print("Running benchmark...")
    Ns, runtime, memory, outputs = benchmark(
        funcs,
        data,
        kernels,
        n_repeat=n_repeat,
        cutoffs=cutoffs,
        drop_outliers=drop_outliers,
        use_gpu_profiler=use_gpu_profiler,
    )
    return Ns, runtime, memory, outputs


def run_pred_benchmark(
    true_kernel,
    funcs,
    kernels,
    yerr=0.3,
    N_N=10,
    maxN=1e5,
    logN_min=1,
    logN_max=5,
    n_repeat=5,
    cutoffs={},  # in M
    drop_outliers=False,
    use_gpu_profiler=False,
    exposure_quantities=None,
):
    runtime = {name: [] for name in funcs}
    memory = {name: [] for name in funcs}
    outputs = {name: [] for name in funcs}

    # Data (N) and test (M)
    Ns = jnp.logspace(logN_min, logN_max, N_N).astype(int)
    Ms = 100 * Ns
    isint = "_int" if exposure_quantities else ""
    for i, (N, M) in enumerate(zip(Ns, Ms), 1):
        skip = True
        for key in cutoffs:
            if M <= cutoffs[key]:
                skip = False
        if skip:
            print(f"  ({i}/{N_N}):  N = {N}, M = {M} -- Skipped (M > all cutoffs)")
            for name in funcs:
                runtime[name].append((jnp.nan, jnp.nan))
                memory[name].append((jnp.nan, jnp.nan))
                outputs[name].append(jnp.nan)
            continue

        print(f"  ({i}/{N_N}):  N = {N}, M = {M}")
        ## Data to condition on/predict from
        datafile = f"data/{N}{isint}.npz"
        if os.path.exists(datafile):
            data = jnp.load(datafile)["arr_0"]
        else:
            data = get_data(
                true_kernel,
                N,
                yerr=yerr,
                exposure_quantities=exposure_quantities,
                save=True,
            )
        if exposure_quantities:
            X_train, y_train, yerr_train = unpack_idata(data)
            t_train, texp_train, instid = X_train
        else:
            t_train, y_train, yerr_train = unpack_data(data)
            X_train = t_train

        # Test grid to predict at
        print("Generating data for benchmarking...")
        dt = 0.1 * (t_train.max() - t_train.min())  # how much to predict/retrodict
        t_test = jnp.linspace(t_train.min() - dt, t_train.max() + dt, M)

        ## Prepare GP objects
        if "QSM" in kernels:
            gp_qs = tinygp.GaussianProcess(kernels["QSM"], X_train, diag=yerr_train**2)
        gp_gp = tinygp.GaussianProcess(kernels["GP"], X_train, diag=yerr_train**2)
        gp_ss = smolgp.GaussianProcess(kernels["SSM"], X_train, diag=yerr_train**2)
        ## Pre-condition those that are compatible with it
        # _, condGPss = gp_ss.condition(y_train)
        ## Pack dict
        # gp = {'SSM': condGPss, 'QSM': gp_qs, 'GP': gp_gp}
        gp = {"SSM": gp_ss, "GP": gp_gp}
        if "QSM" in kernels:
            gp["QSM"] = gp_qs

        _, t, m, o = benchmark(
            funcs,
            [t_test],
            gp,
            y_train,
            n_repeat=n_repeat,
            cutoffs=cutoffs,
            drop_outliers=drop_outliers,
            use_gpu_profiler=use_gpu_profiler,
        )
        for name in funcs:
            if N <= maxN:
                runtime[name].append(t[name][0])
                memory[name].append(m[name][0])
                outputs[name].append(o[name][0])
            else:
                runtime[name].append((jnp.nan, jnp.nan))
                memory[name].append((jnp.nan, jnp.nan))
                outputs[name].append(jnp.nan)

    return Ns, runtime, memory, outputs



from __future__ import annotations

import jax
import jax.numpy as jnp


def RTSSmoother(kernel, X, kalman_results):
    """
    Wrapper for RTS smoother

    Parameters:
        kernel: StateSpaceModel kernel
        X: data coordinates, e.g. time or (time, texp, instid)
        kalman_results: output from Kalman filter (m_filtered, P_filtered, m_predicted, P_predicted)

    Returns:
        m_smooth: smoothed means
        P_smooth: smoothed covariances
    """
    A = kernel.transition_matrix
    t = kernel.coord_to_sortable(X)
    return rts_smoother(A, t, *kalman_results)


@jax.jit
def rts_smoother(A, t, m_filtered, P_filtered, m_predicted, P_predicted):
    """
    Jax implementation of the Rauch-Tung-Striebel (RTS) smoothing algorithm

    See Theorem 8.2 (pdf page 156) in "Bayesian Filtering and Smoothing"
    by Simo S{\"a}rkk{\"a} for detailed description of the algorithm and notation.
    """
    N = len(t)  # number of data points

    def step(carry, k):
        """
        Routine for a single step of the RTS smoother

        Parameters:
            carry: (m_next, P_next) - next state and covariance
            k: index of the current time step

            Recall we are iterating backwards, so _next is k+1

        Returns:
        - Smoothed state (m_k_hat) and covariance (P_k_hat) to carry to next iteration
        - Full output for completed scan (m_k_hat, P_k_hat)
        """

        # Outputs from Kalman filter, unpacked for notational consistency
        m_k = m_filtered[k]
        P_k = P_filtered[k]
        m_pred_next = m_predicted[k + 1]  # has superscript minus
        P_pred_next = P_predicted[k + 1]  # has superscript minus

        # Unpack state and covariance from last iteration
        m_hat_next, P_hat_next = carry

        # Time-lag between states
        Delta_k = t[k + 1] - t[k]

        # Transition matrix
        A_k = A(0, Delta_k)

        # Compute smoothing gain
        # P_pred_next_inv = jnp.linalg.inv(P_pred_next)
        # G_k = P_k @ A_k.T @ P_pred_next_inv # smoothing gain
        G_k = jnp.linalg.solve(P_pred_next.T, (P_k @ A_k.T).T).T  # more stable

        # Update state and covariance
        m_hat_k = m_k + G_k @ (m_hat_next - m_pred_next)
        P_hat_k = P_k + G_k @ (P_hat_next - P_pred_next) @ G_k.T

        return (m_hat_k, P_hat_k), (m_hat_k, P_hat_k)

    # Start smoothing from final filtered state
    init_carry = (m_filtered[-1], P_filtered[-1])

    # Run backward from N-2 down to 0
    _, outputs = jax.lax.scan(step, init_carry, jnp.arange(N - 2, -1, -1))
    m_smooth_reversed, P_smooth_reversed = outputs

    # Reverse outputs to match time order
    m_smooth = jnp.vstack([m_smooth_reversed[::-1], m_filtered[-1][None, :]])
    P_smooth = jnp.vstack([P_smooth_reversed[::-1], P_filtered[-1][None, :, :]])
    return m_smooth, P_smooth
